from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    Ethnicity, HospitalServicesForUrologyCare, MorbidObesity, OfficeVisit, OncAdministrativeSex,
    Payer, Race, UrinaryRetention)


class ClinicalQualityMeasure771v2(ClinicalQualityMeasure):
    """
    Urinary Symptom Score Change 6-12 Months After Diagnosis of Benign Prostatic Hyperplasia

    Description: Percentage of patients with an office visit within the measurement period and with
    a new diagnosis of clinically significant Benign Prostatic Hyperplasia who have International
    Prostate Symptoms Score (IPSS) or American Urological Association (AUA) Symptom Index (SI)
    documented at time of diagnosis and again 6-12 months later with an improvement of 3 points

    Definition: Clinically significant Benign Prostatic Hyperplasia (BPH) is BPH with lower urinary
    tract symptoms (LUTS).
    IPSS - International Prostate Symptom Score
    AUA-SI - American Urological Association-Symptom Index
    QOL - Quality of Life score

    Rationale: Benign prostatic hyperplasia (BPH) is one of the most common conditions affecting
    older men, with a prevalence of 50% by age 60 years and 90% by the ninth decade of life
    (Medina, Parra, & Moore, 1999). The enlarged gland had been proposed to contribute to the
    overall lower urinary tract symptoms (LUTS) complex (McVary et al., 2014). Although LUTS
    secondary to BPH is not often a life-threatening condition, the impact of LUTS/BPH on quality
    of life can be significant (Wei, Calhoun, & Jacobsen, 2005). The American Urological
    Association Symptom Index (AUA-SI) and the International Prostate Symptom Score (IPSS) were
    developed to measure outcomes in studies of different treatments for BPH (Wuerstle et al.,
    2011). The IPSS uses the same questions as the AUA-SI, but also adds a disease-specific quality
    of life question (OLeary, 2005). The IPSS was adopted in 1993 by the World Health Organization.
    It is a reproducible, validated index designed to determine disease severity and response to
    therapy (D’Silva, Dahm, & Wong, 2014). It includes 3 storage symptom questions (frequency,
    nocturia, urgency) and four voiding symptoms (feeling of incomplete emptying, intermittency,
    straining, and a weak stream) as well as a Bother question: If you were to spend the rest of
    your life with your urinary condition just the way it is now, how would you feel about that? A
    three-point improvement in the score is considered meaningful (McVary et al., 2014).

    Guidance: The IPSS is inclusive of the symptom index score and the quality of life score. The
    AUA-SI is the symptom index score alone and must be added to the QOL score. The AUA-SI with the
    QOL equals the IPSS. Both of these are the urinary symptom score.

    The patient must have a urinary symptom score (USS) within 1 month after initial diagnosis. If
    more than one USS in the initial one month, then the first USS counts. The patient must have a
    USS again at 6-12 months after the initial diagnosis and if more than one USS in this time
    frame, then the last USS counts.

    Hospitalization within 30 days of Initial BPH Diagnosis refers to a 30-day period between the
    end of the hospitalization and the clinical office setting BPH diagnosis. This is due to
    aggravating factors from hospitalization, such as bed rest, medications, surgery, and altered
    body functions.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms771v2
    """

    title = 'Urinary Symptom Score Change 6-12 Months After Diagnosis of Benign Prostatic Hyperplasia'

    identifiers = ['CMS771v2']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'Large Urology Group Practice Association (LUGPA)',
    ]

    references = [
        'D\'Silva, K. A., Dahm, P., & Wong, C. L. (2014). Does this man with lower urinary tract symptoms have bladder outlet obstruction? The Rational Clinical Examination: A systematic review. Journal of the American Medical Association, 312(5), 535-542. Retrieved from https://www.ncbi.nlm.nih.gov/pubmed/25096693',
        'McVary, K. T., Roehrborn, C. G., Avins, A. L., et al. (2014). Management of benign prostatic hyperplasia. Retrieved from http://www.auanet.org/guidelines/benign-prostatic-hyperplasia-(2010-reviewed-and-validity-confirmed-2014)',
        'Medina, J. J., Parra, R. O., & Moore, R. G. (1999). Benign prostatic hyperplasia (the aging prostate). Medical Clinics of North America, 83(5), 1213-1229. Retrieved from http://www.sciencedirect.com/science/article/pii/S0025712505701590',
        'O’Leary, M. P. (2005). Validity of the Bother Score in the evaluation and treatment of symptomatic benign prostatic hyperplasia. Reviews in Urology, 7(1), 1-10. Retrieved from https://www.ncbi.nlm.nih.gov/pmc/articles/PMC1477553/',
        'Wei, J., Calhoun, E., & Jacobsen, S. (2005). Urologic diseases in America project: Benign prostatic hyperplasia. Journal of Urology, 173(4), 1256-1261. Retrieved from https://www.ncbi.nlm.nih.gov/pubmed/15758764',
        'Wuerstle, M. C., Van Den Eeden, S. K., Poon, K. T., et al. (2011). Contribution of common medications to lower urinary tract symptoms in men. Archives of Internal Medicine, 171(18), 1680-1682. Retrieved from https://www.ncbi.nlm.nih.gov/pubmed/21987200',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Male patients with an initial diagnosis of benign prostatic hyperplasia
        6 months prior to, or during the measurement period, and a urinary symptom score assessment
        within 1 month of initial diagnosis and a follow-up urinary symptom score assessment within
        6-12 months, who had a qualifying visit during the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Patients with urinary retention that starts within 1 year of initial BPH
        diagnosis.
        Patients with an initial BPH diagnosis that starts during, or within 30 days of
        hospitalization.
        Patients with a diagnosis of morbid obesity, or with a BMI Exam >40 before the follow up
        urinary symptom score.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients with a documented improvement of at least 3 points in their urinary
        symptom score during the measurement period

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The symptoms of BPH are LUTS symptoms. There are other disorders
        with similar symptoms that need to be excluded. History, physical examination, and testing
        are required prior to a diagnosis of BPH. IPSS by itself is not a reliable diagnostic tool
        for LUTS suggestive of BPH, but serves as a quantitative measure of LUTS after the
        diagnosis is established (D’Silva, Dahm, & Wong, 2014). Medical and surgical interventions
        for BPH recommend a follow up IPSS evaluation to determine effectiveness of treatment. IPSS
        should be evaluated at the time of diagnosis and after definitive treatment.
        """
        pass



from canvas_workflow_kit.intervention import Intervention
from canvas_workflow_kit.patient import Patient
from canvas_workflow_kit.recommendation import (
    AllergyRecommendation,
    DiagnoseRecommendation,
    FollowUpRecommendation,
    HyperlinkRecommendation,
    ImagingRecommendation,
    ImmunizationRecommendation,
    InstructionRecommendation,
    InterviewRecommendation,
    LabRecommendation,
    PlanRecommendation,
    PrescribeRecommendation,
    Recommendation,
    ReferRecommendation,
    TaskRecommendation,
    VitalSignRecommendation
)
from canvas_workflow_kit.value_set.v2018 import (
    AboveNormalFollowUp,
    CtColonography,
    Diabetes,
    FitDna,
    Hba1CLaboratoryTest,
    InfluenzaVaccine_1254,
    LimitedLifeExpectancy,
    TobaccoUseCessationPharmacotherapy,
    TobaccoUseScreening
)
from canvas_workflow_kit.value_set.value_set import ValueSet

from .base import SDKBaseTest


class RecommendationTest(SDKBaseTest):

    def test___init__(self):
        rec = Recommendation(
            key='KEY-ID',
            rank=123,
            button='ACT',
            title='Do Something',
            narrative="I'm not sure what, but do something!")
        self.assertTrue(isinstance(rec, Recommendation))
        self.assertEqual(rec.title, 'Do Something')
        self.assertEqual(rec.narrative, "I'm not sure what, but do something!")
        self.assertEqual(rec.command, {})
        self.assertEqual(rec.key, 'KEY-ID')
        self.assertEqual(rec.rank, 123)
        self.assertEqual(rec.button, 'ACT')

    def test_command_filter(self):
        rec = Recommendation(
            key='KEY-ID', rank=123, button='ACT', title='Something', narrative='Do something!')

        class UnitTestReco1(ValueSet):
            OID = '1.2.3.4'
            VALUE_SET_NAME = 'Reco A for the unit tests'
            CPT = {'abc', '123'}
            SNOMEDCT = {'12345', '23456', '34567'}

        class UnitTestReco2(ValueSet):
            OID = '5.2.3.4'
            VALUE_SET_NAME = 'Reco B for the unit tests'
            CPT = {'xbc', '423'}
            SNOMEDCT = {'612345', '73456', '84567'}

        expected = {
            'type': 'atype',
            'filter': {
                'coding': [{
                    'system': 'cpt',
                    'code': ['123', 'abc', 'xbc', '423']
                }, {
                    'system': 'snomedct',
                    'code': ['34567', '12345', '23456', '612345', '73456', '84567']
                }]
            }
        }  # yapf: disable
        result = rec.command_filter('atype', [UnitTestReco1, UnitTestReco2])
        HelperCommand().equal_commands(expected, result)


class AllergyRecommendationTest(SDKBaseTest):
    def test__init__(self):
        class AllergyTestValueSet(ValueSet):
            VALUE_SET_NAME = 'Enterococcus faecalis'
            FDB = {'1226'}

        rec = AllergyRecommendation(
            key='ALLERGY', rank=1, button='Allergy', allergy=AllergyTestValueSet,
            narrative='rash and hives', context={'severity': 'mild', 'onset_date': '2020-01-01'})
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'ALLERGY')
        self.assertEqual(rec.rank, 1)
        self.assertEqual(rec.button, 'Allergy')
        self.assertEqual(rec.title, 'Enterococcus faecalis')
        self.assertEqual(rec.narrative, 'rash and hives')
        self.assertEqual(rec.context['severity'], 'mild')
        self.assertEqual(rec.context['onset_date'], '2020-01-01')

        expected = {
            'type': 'allergy',
            'filter': {
                'coding': [{
                    'system': 'fdb',
                    'code': ['1226']
                }]
            }
        }
        result = rec.command
        HelperCommand().equal_commands(expected, result)


class DiagnoseRecommendationTest(SDKBaseTest):
    def test__init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        class DiagnoseTestValueSet(ValueSet):
            VALUE_SET_NAME = 'Senile purpura'
            ICD10CM = {'D692'}
            INTERNAL = {'PAT.QUES.8'}

        rec = DiagnoseRecommendation(
            key='DIAGNOSE', rank=1, button='Diagnose', patient=patient,
            condition=DiagnoseTestValueSet)
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'DIAGNOSE')
        self.assertEqual(rec.rank, 1)
        self.assertEqual(rec.button, 'Diagnose')
        self.assertEqual(rec.title, 'Senile purpura')
        self.assertEqual(rec.narrative, 'Consider diagnosing Darey with Senile purpura')

        expected = {
            'type': 'diagnose',
            'filter': {
                'coding': [{
                    'system': 'icd10cm',
                    'code': ['D692']
                }, {
                    'system': 'internal',
                    'code': ['PAT.QUES.8']
                }]
            }
        }
        result = rec.command
        HelperCommand().equal_commands(expected, result)


class FollowUpRecommendationTest(SDKBaseTest):
    def test__init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        rec = FollowUpRecommendation(
            key='FOLLOWUP', rank=1, button='Follow Up', patient=patient,
            title='Request follow up office visit on 3/5/22',
            context={'requested_appointment_type': 'office', 'requested_date': '2022-03-05',
                     'reason_for_visit': 'covid vaccine booster',
                     'internal_comment': 'make sure he brings his covid vaccination card'})
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'FOLLOWUP')
        self.assertEqual(rec.rank, 1)
        self.assertEqual(rec.button, 'Follow Up')
        self.assertEqual(rec.title, 'Request follow up office visit on 3/5/22')
        self.assertEqual(rec.narrative, 'Darey needs a follow-up appointment')
        self.assertEqual(rec.context['requested_appointment_type'], 'office')
        self.assertEqual(rec.context['requested_date'], '2022-03-05')
        self.assertEqual(
            rec.context['internal_comment'],
            'make sure he brings his covid vaccination card')
        self.assertEqual(rec.command, {'type': 'followUp'})


class ImagingRecommendationTest(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        rec = ImagingRecommendation(
            key='KEY-ID', rank=123, button='ACT', patient=patient, imaging=CtColonography,
            context={'foo': 'bar'})
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'KEY-ID')
        self.assertEqual(rec.rank, 123)
        self.assertEqual(rec.button, 'ACT')

        expected = 'Darey should be ordered CT Colonography.'
        self.assertEqual(expected, rec.narrative)

        expected = 'Order CT Colonography'
        self.assertEqual(expected, rec.title)

        self.assertEqual(rec.context['foo'], 'bar')

        expected = {
            'type': 'imagingOrder',
            'filter': {
                'coding': [{
                    'system': 'cpt',
                    'code': ['74263']
                }, {
                    'system': 'snomedct',
                    'code': ['418714002']
                }]
            }
        }
        result = rec.command
        HelperCommand().equal_commands(expected, result)


class ReferRecommendationTest(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        rec = ReferRecommendation(key='KEY-ID', rank=123, button='ACT',
                                  patient=patient, referral=FitDna, context={'foo': 'bar'})
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'KEY-ID')
        self.assertEqual(rec.rank, 123)
        self.assertEqual(rec.button, 'ACT')

        expected = 'Darey should be referred for FIT DNA.'
        self.assertEqual(expected, rec.narrative)

        expected = 'Refer FIT DNA'
        self.assertEqual(expected, rec.title)

        self.assertEqual(rec.context['foo'], 'bar')

        expected = {
            'type': 'refer',
            'filter': {
                'coding': [{
                    'system': 'loinc',
                    'code': ['77354-9', '77353-1']
                }]
            }
        }
        result = rec.command
        HelperCommand().equal_commands(expected, result)


class LabRecommendationTest(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        rec = LabRecommendation(
            key='KEY-ID',
            rank=123,
            button='ACT',
            patient=patient,
            lab=Hba1CLaboratoryTest,
            condition=Diabetes,
            context={'foo': 'bar'})
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'KEY-ID')
        self.assertEqual(rec.rank, 123)
        self.assertEqual(rec.button, 'ACT')

        expected = 'Darey has Diabetes and a HbA1c Laboratory Test is recommended.'
        self.assertEqual(expected, rec.narrative)

        expected = 'Order HbA1c Laboratory Test'
        self.assertEqual(expected, rec.title)

        self.assertEqual(rec.context['foo'], 'bar')

        expected = {
            'type': 'labOrder',
            'filter': {
                'coding': [{
                    'system': 'loinc',
                    'code': ['4548-4', '17856-6', '4549-2']
                }]
            }
        }
        result = rec.command
        HelperCommand().equal_commands(expected, result)


class ImmunizationRecommendationTest(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        rec = ImmunizationRecommendation(
            key='KEY-ID',
            rank=123,
            button='ACT',
            patient=patient,
            immunization=InfluenzaVaccine_1254,
            context={'foo': 'bar'})
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'KEY-ID')
        self.assertEqual(rec.rank, 123)
        self.assertEqual(rec.button, 'ACT')

        expected = 'Darey should get an Influenza Vaccine.'
        self.assertEqual(expected, rec.narrative)

        expected = 'Order Influenza Vaccine'
        self.assertEqual(expected, rec.title)

        self.assertEqual(rec.context['foo'], 'bar')

        expected = {
            'type': 'immunization',
            'filter': {
                'coding': [{
                    'system': 'cvx',
                    'code': [
                        '88', '155', '168', '171', '150', '158', '135', '144', '166', '141', '161',
                        '153', '140'
                    ]
                }]
            }
        }
        result = rec.command
        HelperCommand().equal_commands(expected, result)


class InstructionRecommendationTest(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        rec = InstructionRecommendation(
            key='KEY-ID', rank=123, button='ACT', patient=patient, instruction=AboveNormalFollowUp,
            context={'foo': 'bar'})
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'KEY-ID')
        self.assertEqual(rec.rank, 123)
        self.assertEqual(rec.button, 'ACT')

        expected = 'Darey should get an Above Normal Follow-up.'
        self.assertEqual(expected, rec.narrative)

        expected = 'Instruct Above Normal Follow-up'
        self.assertEqual(expected, rec.title)

        self.assertEqual(rec.context['foo'], 'bar')

        expected = {
            'type': 'instruct',
            'filter': {
                'coding': [{
                    'system': 'cpt',
                    'code': [
                        '97804', '43888', '97802', '43774', '43772', '98960', '43845', '43842',
                        '99401', '43843', '43644', '43846', '43770', '43773', '97803', '43659',
                        '43645', '43847', '99402', '43771', '43886', '99078', '43848'
                    ]
                }, {
                    'system': 'hcpcs',
                    'code': [
                        'G0447', 'G0271', 'S9449', 'S9451', 'S9470', 'G0473', 'S9452', 'G0270'
                    ]
                }, {
                    'system': 'icd10cm',
                    'code': ['Z713']
                }, {
                    'system': 'snomedct',
                    'code': [
                        '386291006', '386292004', '361231003', '370847001', '413315001',
                        '304549008', '424753004', '386464006', '307818003', '386373004',
                        '386463000', '418995006', '443288003', '410177006'
                    ]
                }]
            }
        }  # yapf: disable
        result = rec.command
        HelperCommand().equal_commands(expected, result)


class InterviewRecommendationTest(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        rec = InterviewRecommendation(
            key='KEY-ID',
            rank=123,
            button='ACT',
            patient=patient,
            questionnaires=[TobaccoUseScreening, LimitedLifeExpectancy],
            context={'foo': 'bar'})
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'KEY-ID')
        self.assertEqual(rec.rank, 123)
        self.assertEqual(rec.button, 'ACT')

        expected = 'Darey should be given a Tobacco Use Screening, Limited Life Expectancy.'
        self.assertEqual(expected, rec.narrative)

        expected = 'Interview Tobacco Use Screening, Limited Life Expectancy'
        self.assertEqual(expected, rec.title)

        self.assertEqual(rec.context['foo'], 'bar')

        expected = {
            'type': 'interview',
            'filter': {
                'coding': [{
                    'system': 'loinc',
                    'code': ['72166-2', '68536-2', '39240-7', '68535-4']
                }, {
                    'system': 'snomedct',
                    'code': ['300936002', '27143004', '162608008', '170969009', '162607003']
                }]
            }
        }  # yapf: disable
        result = rec.command
        HelperCommand().equal_commands(expected, result)


class PrescribeRecommendationTest(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        rec = PrescribeRecommendation(
            key='KEY-ID',
            rank=123,
            button='ACT',
            patient=patient,
            prescription=TobaccoUseCessationPharmacotherapy,
            context={'foo': 'bar'})
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'KEY-ID')
        self.assertEqual(rec.rank, 123)
        self.assertEqual(rec.button, 'ACT')

        expected = 'Darey should be prescribed Tobacco Use Cessation Pharmacotherapy.'
        self.assertEqual(expected, rec.narrative)

        expected = 'Prescribe Tobacco Use Cessation Pharmacotherapy'
        self.assertEqual(expected, rec.title)

        self.assertEqual(rec.context['foo'], 'bar')

        expected = {
            'type': 'prescribe',
            'filter': {
                'coding': [{
                    'system': 'rxnorm',
                    'code': [
                        '892244', '993550', '151226', '198045', '199889', '1551468', '993691',
                        '198047', '1801289', '1232585', '317136', '998675', '993518', '993536',
                        '1797886', '993567', '198046', '359817', '998679', '993541', '993557',
                        '359818', '199283', '198029', '311975', '250983', '749788', '198030',
                        '749289', '998671', '205315', '993687', '205316', '636671', '993681',
                        '993503', '199890', '198031', '199888', '312036', '636676', '314119'
                    ]  # yapf: disable
                }]
            }
        }
        result = rec.command
        HelperCommand().equal_commands(expected, result)


class VitalSignRecommendationTest(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        rec = VitalSignRecommendation(key='KEY-ID', rank=123, button='ACT',
                                      patient=patient, context={'foo': 'bar'})
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'KEY-ID')
        self.assertEqual(rec.rank, 123)
        self.assertEqual(rec.button, 'ACT')
        self.assertEqual('Darey should vital signs collected.', rec.narrative)
        self.assertEqual('Collect vitals', rec.title)
        self.assertEqual({'type': 'vitalSign'}, rec.command)
        self.assertEqual(rec.context['foo'], 'bar')


class PlanRecommendationTest(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        rec = PlanRecommendation(key='KEY-ID', rank=123, button='ACT',
                                 patient=patient, context={'foo': 'bar'})
        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'KEY-ID')
        self.assertEqual(rec.rank, 123)
        self.assertEqual(rec.button, 'ACT')
        self.assertEqual('Darey should have a plan.', rec.narrative)
        self.assertEqual('Make a plan', rec.title)
        self.assertEqual({'type': 'plan'}, rec.command)
        self.assertEqual(rec.context['foo'], 'bar')


class TaskRecommendationTest(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')
        self.assertIsInstance(patient, Patient)

        rec = TaskRecommendation(key='KEY-ID', rank=123, title='TITLE',
                                 button='ACT', patient=patient, context={'foo': 'bar'})

        self.assertIsInstance(rec, Recommendation)
        self.assertEqual(rec.key, 'KEY-ID')
        self.assertEqual(rec.title, 'TITLE')
        self.assertEqual(rec.rank, 123)
        self.assertEqual(rec.button, 'ACT')
        self.assertIsNone(rec.narrative)
        self.assertEqual({'type': 'task'}, rec.command)
        self.assertEqual(rec.context['foo'], 'bar')


class HyperlinkRecommendationTest(SDKBaseTest):
    def test___init__(self):
        rec = HyperlinkRecommendation(
            key='KEY!', rank=4, button='Try this', href='https://www.google.com/search?q=broccoli',
            title='Something needs to be done')
        self.assertIsInstance(rec, Recommendation)
        self.assertIsInstance(rec, Intervention)
        self.assertEqual(rec.key, 'KEY!')
        self.assertEqual(rec.title, 'Something needs to be done')
        self.assertEqual(rec.rank, 4)
        self.assertEqual(rec.button, 'Try this')
        self.assertEqual(rec.href, 'https://www.google.com/search?q=broccoli')


# --- helpers ---
class HelperCommand(SDKBaseTest):

    def equal_commands(self, expected: dict, result: dict):
        self.assertTrue(('type' in result) and (result['type'] == expected['type']))
        self.assertTrue(('filter' in result) and ('coding' in result['filter']))
        self.assertEqual(len(expected['filter']['coding']), len(result['filter']['coding']))

        for l_coding in expected['filter']['coding']:
            tmp = False
            for r_coding in result['filter']['coding']:
                if l_coding['system'] == r_coding['system']:
                    tmp = self.are_dicts_same(l_coding, r_coding)
                    break
            if not tmp:
                self.assertTrue(False, 'filter.coding with system: %s' % l_coding['system'])

    def are_dicts_same(self, left: dict, right: dict) -> bool:
        for key, val in left.items():
            result = False
            if key in right and isinstance(right[key], type(val)):
                if isinstance(val, dict):
                    result = self.are_dicts_same(val, right[key])
                elif isinstance(val, list):
                    result = True if sorted(val) == sorted(right[key]) else False
                else:
                    result = True if val == right[key] else False
            if not result:
                return False
        return True



from ..value_set import ValueSet


class BestCorrectedVisualAcuityExamUsingSnellenChart(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent best corrected visual acuity exams using a Snellen chart.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) category related to Physical Exam.

    **Inclusion Criteria:** Includes only relevant concepts associated with best corrected visual acuity exams using a Snellen chart.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1560'
    VALUE_SET_NAME = 'Best Corrected Visual Acuity Exam Using Snellen Chart'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '79880-1',
        '79881-9'
    }

class BmiPercentile(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent BMI percentile.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Physical Exam, Performed.

    **Inclusion Criteria:** Includes only relevant concepts associated with identifying BMI percentile as a result. This is a grouping of LOINC codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.121.12.1012'
    VALUE_SET_NAME = 'BMI percentile'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '59574-4',
        '59575-1',
        '59576-9'
    }

class HeartRate(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent heart rate.

    **Data Element Scope:** This value set may use Quality Data Model (QDM) category related to Physical Exam.

    **Inclusion Criteria:** Includes only relevant concepts associated with obtaining heart rate.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1176'
    VALUE_SET_NAME = 'Heart Rate'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '68999-2',
        '69000-8',
        '69001-6',
        '8867-4'
    }

class Height(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent a patient's height has been measured or observed.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Physical Exam, Performed.

    **Inclusion Criteria:** Includes only relevant concepts associated with identifying if a patient's height has been measured or observed. This is a grouping of LOINC codes.

    **Exclusion Criteria:** Excludes codes that are expressed as a percentile, stated by a patient, post-partum or estimated.
    """

    OID = '2.16.840.1.113883.3.464.1003.121.12.1014'
    VALUE_SET_NAME = 'Height'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '3137-7',
        '8302-2',
        '8306-3',
        '8307-1',
        '8308-9'
    }

class RetinalOrDilatedEyeExam(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent patients who have received a retinal or dilated eye exam.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Physical Exam, Performed.

    **Inclusion Criteria:** Includes only relevant concepts associated with procedural codes that indicate a retinal and dilated eye exam took place.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.115.12.1088'
    VALUE_SET_NAME = 'Retinal or Dilated Eye Exam'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '252779009',
        '252780007',
        '252781006',
        '252782004',
        '252783009',
        '252784003',
        '252788000',
        '252789008',
        '252790004',
        '274795007',
        '274798009',
        '308110009',
        '314971001',
        '314972008',
        '410451008',
        '410452001',
        '410453006',
        '410455004',
        '420213007',
        '425816006',
        '427478009',
        '6615001',
        '722161008'
    }

class Weight(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent that a patient's weight has been measured.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Physical Exam, Performed.

    **Inclusion Criteria:** Includes only relevant concepts associated with identifying a patient's weight has been measured. This is a grouping of LOINC codes.

    **Exclusion Criteria:** Excludes codes for patient's ideal body weight, estimated body weight and measured/calculated body fat.
    """

    OID = '2.16.840.1.113883.3.464.1003.121.12.1015'
    VALUE_SET_NAME = 'Weight'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '18833-4',
        '29463-7',
        '3141-9',
        '3142-7',
        '8341-0',
        '8349-3',
        '8350-1',
        '8351-9'
    }


from canvas_workflow_kit.value_set.value_set import ValueSet


class StructuredAssessmentQuestionnaireMoodDisorder(ValueSet):

    VALUE_SET_NAME = 'SCAN Mood Disorder Structured Assessment Questionnaire'

    INTERNAL = {'COGN_SCAN_MOOD_DISORDER'}


class QuestionnairePVQ(ValueSet):

    VALUE_SET_NAME = 'SCAN Pre-visit Questionnaire'

    SNOMEDCT = {'456201000124103'}


class StructuredAssessmentQuestionnaireUrinarySymptoms(ValueSet):

    VALUE_SET_NAME = 'SCAN Urinary Symptoms Structured Assessment Questionnaire'

    INTERNAL = {'MEDI_SCAN_URINARY_SYMPTOMS_ASSESSMENT'}

class StructuredAssessmentQuestionnaireCognitiveDisorders(ValueSet):

    VALUE_SET_NAME = 'SCAN Cognitive Disorders Structured Assessment Questionnaire'

    INTERNAL = {'COGN_SCAN_COGNITIVE_DISORDERS_ASSESSMENT'}


class QuestionnaireHealthCHECReferrals(ValueSet):

    VALUE_SET_NAME = 'SCAN HealthCHEC Referrals Questionnaire'

    INTERNAL = {'COOR_183522000'}


class QuestionnaireComprehensiveGeriatricAssessment(ValueSet):

    VALUE_SET_NAME = 'SCAN Comprehensive Geriatric Assessment Questionnaire'

    SNOMEDCT = {'CGA_711013002'}



from ..value_set import ValueSet


class MedicalReason(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent medical reasons for when a patient does not receive a therapy or service.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) attribute related to Negation Rationale.

    **Inclusion Criteria:** Includes only relevant concepts associated with medically relevant reasons for not receiving a therapy or service.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1007'
    VALUE_SET_NAME = 'Medical Reason'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '183932001',
        '183964008',
        '183966005',
        '266721009',
        '269191009',
        '31438003',
        '35688006',
        '397745006',
        '407563006',
        '410534003',
        '410536001',
        '416098002',
        '428119001',
        '59037007',
        '62014003',
        '79899007'
    }

class PatientDeclined(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent a reason for patient refusal.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) attribute related to Negation Rationale.

    **Inclusion Criteria:** Includes only relevant concepts associated with identifying a patient's reason for refusing treatment.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.600.791'
    VALUE_SET_NAME = 'Patient Declined'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '105480006',
        '183944003',
        '183945002',
        '413310006',
        '413311005',
        '413312003'
    }

class PatientDeclined_1582(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent a reason for patient refusal.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) attribute related to Negation Rationale.

    **Inclusion Criteria:** Includes only relevant concepts associated with identifying a patient's reason for refusing treatment.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1582'
    VALUE_SET_NAME = 'Patient Declined'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '105480006',
        '183944003',
        '183945002',
        '413310006',
        '413311005',
        '413312003'
    }

class PatientReason(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent patient-specific reasons for when a patient does not receive a therapy or service.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) attribute related to Negation Rationale.

    **Inclusion Criteria:** Includes only relevant concepts associated with patient-specific reasons for not receiving a therapy or service.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1008'
    VALUE_SET_NAME = 'Patient Reason'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '105480006',
        '160932005',
        '160934006',
        '182890002',
        '182895007',
        '182897004',
        '182900006',
        '182902003',
        '183944003',
        '183945002',
        '184081006',
        '185479006',
        '185481008',
        '224187001',
        '225928004',
        '266710000',
        '266966009',
        '275694009',
        '275936005',
        '281399006',
        '310343007',
        '373787003',
        '406149000',
        '408367005',
        '413310006',
        '413311005',
        '413312003',
        '416432009',
        '423656007',
        '424739004',
        '443390004',
        '713247000'
    }

class SystemReason(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent system-specific reasons for when a patient does not receive a therapy or service.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) attribute related to Negation Rationale.

    **Inclusion Criteria:** Includes only relevant concepts associated with system-specific reasons for not receiving a therapy or service.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1009'
    VALUE_SET_NAME = 'System Reason'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '107724000',
        '182856006',
        '182857002',
        '185335007',
        '224194003',
        '224198000',
        '224199008',
        '242990004',
        '266756008',
        '270459005',
        '309017000',
        '309846006',
        '419808006',
        '424553001'
    }


from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    EncounterInpatient, Ethnicity, FractureLowerBody, HospiceCareAmbulatory,
    KneeInjuryAndOsteoarthritisOutcomeScoreForJointReplacementKoosjr, OfficeVisit,
    OncAdministrativeSex, OutpatientConsultation, Payer, Race)


class ClinicalQualityMeasure66v9(ClinicalQualityMeasure):
    """
    Functional Status Assessment for Total Knee Replacement

    Description: Percentage of patients 18 years of age and older who received an elective primary
    total knee arthroplasty (TKA) and completed a functional status assessment within 90 days prior
    to the surgery and in the 270-365 days after the surgery

    Definition: None

    Rationale: Total knee arthroplasties (TKAs) are common surgical procedures for addressing knee
    pain and functional impairment, primarily caused by osteoarthritis. From 2008 to 2010, TKAs
    were the most common procedure for adults age 45 and older. From 2000 to 2010, physicians
    performed 5.2 million TKAs, with 693,400 procedures performed in 2010 alone (Williams, Wolford,
    & Bercovitz, 2015). Although TKAs were introduced as a procedure for older adults, the mean age
    of patients undergoing TKA is decreasing. In 2010, the mean age for those undergoing TKA was
    66.2, a 3.9 percent decrease from 68.9 in 2000. Kurtz et al. (2009) projected that patients
    younger than 65 would account for 55 percent of TKAs by 2030. This growth in knee surgeries for
    younger patients is significant because they often require more expensive joint arthroplasties
    that will better withstand wear caused by physical activity (Bozic et al., 2006).

    This measure evaluates whether patients complete a patient-reported functional status
    assessment (FSA) before and after a TKA. Measuring functional status for patients undergoing
    total knee replacement permits longitudinal assessment - from the patient's perspective - of
    the impact of surgical intervention on pain, physical function, as well as health-related
    quality of life (Rothrock et al., 2010).

    Guidance: The same functional status assessment (FSA) instrument must be used for the initial
    and follow-up assessment.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms66v9
    """

    title = 'Functional Status Assessment for Total Knee Replacement'

    identifiers = ['CMS66v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'American Association of Orthopaedic Surgeons, the American Joint Replacement Registry, The Hip Society, The Knee Society, & the American Association of Hip and Knee Surgeons. (2015). Patient-reported outcomes summit for total joint arthroplasty report. Retrieved from http://www.aahks.org/wp-content/uploads/2018/08/comment-letter-pro-summit-09082015.pdf',
        'Bozic, K. J., Morsehed, S., Silverstein, M. D., et al. (2006). Use of cost-effectiveness analysis to evaluate new technologies in orthopaedics: The case of alternative bearing surfaces in total hip arthroplasty. Journal of Bone and Joint Surgery, American Volume, 88(4), 706-714.',
        'Kurtz, S. M., Lau, E., Ong, K., et al. (2009). Future young patient demand for primary and revision joint arthroplasty: National projections from 2010 to 2030. Clinical Orthopaedics and Related Research, 467(10), 2606-2612.',
        'Rothrock, N. E., Hays, R. D., Spritzer, K., et al. (2010). Relative to the general U.S. population, chronic diseases are associated with poorer health-related quality of life as measured by the Patient-Reported Outcomes Measurement Information System (PROMIS). Journal of Clinical Epidemiology, 63(11), 1195-1204.',
        'Williams, S. N., Wolford, M. L., & Bercovitz, A. (2015). Hospitalization for total knee replacement among inpatients aged 45 and over: United States, 2000-2010. NCHS Data Brief No. 210. Hyattsville, MD: National Center for Health Statistics.',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients 19 years of age and older who had a primary total knee
        arthroplasty (TKA) in the year prior to the measurement period and who had an outpatient
        encounter during the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Patients with two or more fractures indicating trauma at the time of the total
        knee arthroplasty or patients with severe cognitive impairment that overlaps the
        measurement period.

        Exclude patients whose hospice care overlaps the measurement period.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients with patient-reported functional status assessment results (i.e.,
        Veterans RAND 12-item health survey [VR-12], Patient-Reported Outcomes Measurement
        Information System [PROMIS]-10 Global Health, Knee Injury and Osteoarthritis Outcome Score
        [KOOS], KOOS Jr.) in the 90 days prior to or on the day of the primary TKA procedure, and
        in the 270 - 365 days after the TKA procedure

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: While there is no clinical guideline recommending that clinicians
        assess patients who are undergoing total knee replacements using patient-reported outcomes
        of function and pain, several clinical specialty societies support the use of a general
        health questionnaire and a disease-specific questionnaire for these patients. In
        particular, they recommend the Veterans RAND 12-item health survey (VR-12) or the Patient-
        Reported Outcomes Measurement Information System [PROMIS]-10-Global as the general health
        questionnaire and the Knee Injury and Osteoarthritis Outcome Score [KOOS], Jr. as the
        disease-specific questionnaire (American Association of Orthopaedic Surgeons, the American
        Joint Replacement Registry, The Hip Society, The Knee Society, & the American Association
        of Hip and Knee Surgeons, 2015).
        """
        pass



from typing import Dict, List, Optional, cast

import arrow

from cached_property import cached_property

from canvas_workflow_kit import events, settings
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ExternallyAwareClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import InterviewRecommendation
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.value_set.customer_value_sets.scan import (
    QuestionnaireHealthCHECReferrals,
    QuestionnairePVQ
)


class CustomerScanComplexCaseManagementReferral(ExternallyAwareClinicalQualityMeasure,
                                                ClinicalQualityMeasure):
    """
    A clinical quality measure (which is a type of protocol) used only by one customer (SCAN).
    More information at https://docs.google.com/document/d/1NW4b3OT_yPeZy_oIKt_DCCeQ0egDaWFE/edit.
    """

    class Meta:
        title = 'SCAN: Complex Case Management Referral'
        version = '2020-10-13v1'
        description = 'A clinical quality measure (which is a type of protocol) used only by SCAN.'
        information = 'https://docs.google.com/document/d/1NW4b3OT_yPeZy_oIKt_DCCeQ0egDaWFE/edit'

        identifiers = ['CustomerScanComplexCaseManagementReferralV1']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        authors = [
            'SCAN',
        ]
        references = ['Written by SCAN']

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_INTERVIEW,
        ]

        funding_source = 'SCAN'

    @classmethod
    def enabled(cls) -> bool:
        return cast(bool, settings.IS_HEALTHCHEC)

    def in_initial_population(self) -> bool:
        return True

    def interview_lookback_timeframe(self) -> Timeframe:
        return Timeframe(start=self.timeframe.end.shift(days=-30), end=self.timeframe.end)

    @cached_property
    def recent_pvqs(self) -> List[Dict]:
        interviews = cast(
            List[Dict],
            self.patient.interviews.find(QuestionnairePVQ).within(
                self.interview_lookback_timeframe()))

        for interview in interviews:
            interview['groupedResponses'] = self.group_question_responses(interview)

        return interviews

    @cached_property
    def last_referral_for_complex_case_management(self) -> Optional[Dict]:
        interviews = cast(
            List[Dict],
            self.patient.interviews.find(QuestionnaireHealthCHECReferrals).within(
                self.interview_lookback_timeframe()))

        if not interviews:
            return None

        # Go through them in order with the most recent one first and working
        # backwards
        # return the first one with a response representing a physical
        # activity intervention
        for interview in sorted(
                interviews, key=lambda i: arrow.get(i['noteTimestamp']), reverse=True):
            for response in interview.get('responses', []):
                if response.get('code') == '698943000':
                    return interview

        # nothing found
        return None

    def last_referral_for_complex_case_management_date(self) -> Optional[arrow.Arrow]:
        if self.last_referral_for_complex_case_management:
            return arrow.get(self.last_referral_for_complex_case_management['noteTimestamp'])
        return None

    def referred_for_complex_case_management_on_or_after(
            self, reference_date: Optional[arrow.Arrow]) -> bool:
        # It can't have been referred after a date that doesn't exist!
        if not reference_date:
            return False

        # If it hasn't been referred, it certainly wasn't referred after the
        # reference date.
        if not self.last_referral_for_complex_case_management_date():
            return False

        return self.last_referral_for_complex_case_management_date() >= reference_date

    def latest_case_complexity_indication_date(self) -> Optional[arrow.Arrow]:
        dates = cast(List[arrow.Arrow],
                     [date for date in self.latest_case_complexity_indicators().values() if date])
        if dates:
            return max(dates)
        return None

    def latest_case_complexity_indicators(self) -> Dict:
        questions_and_responses_that_indicate_case_complexity: Dict[str, str] = {
            '123978lijsdf293': '123978lijsdf295',
            '123978lijsdf344': '123978lijsdf345',
            '123978lijsdf324': '123978lijsdf326',
            '123978lijsdf350': '123978lijsdf351',
        }

        dates_of_indication: Dict[str, Optional[arrow.Arrow]] = {
            '123978lijsdf293': None,
            '123978lijsdf344': None,
            '123978lijsdf324': None,
            '123978lijsdf350': None,
        }

        # Go through all pvq interviews within the past 30 days in order from
        # the earliest.
        for interview in sorted(self.recent_pvqs, key=lambda i: arrow.get(i['noteTimestamp'])):
            # Check the currently observed interview for each of the trigger
            # questions.
            for question_coding, response_coding in questions_and_responses_that_indicate_case_complexity.items(  # noqa: E501
            ):  # noqa: E501
                question_responses = interview.get('groupedResponses', {}).get(question_coding, [])
                # Check if the question was responded to.
                if question_responses:
                    # Since it was responded to, we consider this a more
                    # updated answer to the question.
                    # If the question's response is one that indicates case
                    # complexity, we'll put this interview's date as the most
                    # recent indication of case complexity. If there are
                    # responses after this, it will get overwritten with the
                    # more recent date.
                    if response_coding in question_responses:
                        dates_of_indication[question_coding] = arrow.get(
                            interview['noteTimestamp'])
                    # If the question's response is not one that indicates
                    # case complexity, we set this question's slot to None,
                    # even if it was previously filled. We are only
                    # considering the most recent response for each question.
                    # If a question is not responded to, the previous response
                    # will remain considered.
                    else:
                        dates_of_indication[question_coding] = None
        return dates_of_indication

    def in_denominator(self) -> bool:
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients who were in hospice care during the measurement year

        Exceptions: None
        """
        if not self.in_initial_population():
            return False

        if self.patient.hospice_within(self.timeframe):
            return False

        return True

    def in_numerator(self) -> bool:
        """
        A patient should be in the numerator if they either:
        1) Have been determined to not need anything addressed (regarding the
           CustomerScanComplexCaseManagementReferral quality measure)
        OR
        2) Have been addressed (regarding the CustomerScanComplexCaseManagementReferral
           quality measure)
        """

        case_complexity_indicated = False
        referred_for_complex_case_management = False

        if self.latest_case_complexity_indication_date():
            case_complexity_indicated = True

        no_case_complexity_indicated = not case_complexity_indicated

        referred_for_complex_case_management = self.referred_for_complex_case_management_on_or_after(  # noqa: E501
            self.latest_case_complexity_indication_date())

        return no_case_complexity_indicated or (case_complexity_indicated and
                                                referred_for_complex_case_management)

    def craft_satisfied_result(self) -> ProtocolResult:
        """
        Creates and returns a ProtocolResult instance that indicates a satisfied result.
        The narrative added either indicates that the patient was referred to complex case
        management or that the patient did not indicate case complexity within a timeframe.
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_SATISFIED

        referred_for_complex_case_management = self.referred_for_complex_case_management_on_or_after(  # noqa: E501
            self.latest_case_complexity_indication_date())

        if referred_for_complex_case_management:
            result.add_narrative(
                f'{self.patient.first_name} was referred for complex case '
                f'management {self.display_date(self.last_referral_for_complex_case_management_date())}.'  # noqa: E501
            )
        else:
            result.add_narrative(
                f'{self.patient.first_name} has not indicated case complexity in the past 30 days.'
            )

        return result

    def craft_unsatisfied_result(self) -> ProtocolResult:
        """
        Creates and returns a ProtocolResult instance that indicates an unsatisfied result.
        Adds an interview recommendation to the protocol result as well as a narrative based on
        the patient's answers.
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_DUE

        narrative_description_from_question_coding = {
            '123978lijsdf293': 'reported feeling unsafe',
            '123978lijsdf344': 'reported medication issues',
            '123978lijsdf324': 'reported not having emergency help',
            '123978lijsdf350': 'reported food insecurity',
        }

        for question_coding, response_date in self.latest_case_complexity_indicators().items():
            if response_date:
                narrative_description = narrative_description_from_question_coding.get(
                    question_coding, 'indicated a case complexity')
                result.add_narrative(f'{self.patient.first_name} {narrative_description} '
                                     f'{self.display_date(response_date)}.')

        result.add_recommendation(
            InterviewRecommendation(
                key='CUSTOMER_SCAN_RECOMMEND_COMPLEX_CASE_MANAGEMENT_REFERRAL',
                rank=1,
                button='Refer',
                patient=self.patient,
                questionnaires=[QuestionnaireHealthCHECReferrals],
                title='Refer for Complex Case Management'))

        return result



import arrow

from unittest import TestCase

from canvas_workflow_kit.patient_recordset import (
    AppointmentRecordSet,
    BillingLineItemRecordSet,
    ConditionRecordSet,
    ConsentRecordSet,
    GroupRecordSet,
    ImagingReportRecordSet,
    ImmunizationRecordSet,
    InstructionRecordSet,
    InterviewRecordSet,
    LabReportRecordSet,
    MedicationRecordSet,
    PatientEventRecordSet,
    PatientPeriodRecordSet,
    PatientRecordSet,
    ProcedureRecordSet,
    ProtocolOverrideRecordSet,
    PrescriptionRecordSet,
    ReferralReportRecordSet,
    TaskRecordSet,
    VitalSignRecordSet
)
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.value_set.value_set import ValueSet


class TestPatientRecordSet(TestCase):

    def test___init__(self):
        tested = self.get_object()
        result = tested.records
        expected = [
            190,
            194,
            193,
            192,
        ]

        self.assertEqual(len(expected), len(result))

        for exp, res in zip(expected, result):
            self.assertEqual(exp, res['id'])

        self.assertEqual(expected, [item['id'] for item in result])

        self.assertEqual('patient__key', PatientRecordSet.API_KEY_FIELD)
        self.assertIsNone(PatientRecordSet.API_UPDATE_FIELD)
        self.assertIsNone(PatientRecordSet.VALID_SYSTEMS)
        self.assertEqual('date', PatientRecordSet.DATE_FIELD)

    def test___str__(self):
        tested = self.get_object()
        result = str(tested)
        expected = '<UnitTestARecordSet: 4 records>'
        self.assertEqual(expected, result)

    def test___repr__(self):
        tested = self.get_object()
        result = repr(tested)
        expected = '<UnitTestARecordSet: 4 records>'
        self.assertEqual(expected, result)

    def test___iter__(self):
        tested = self.get_object()
        expected = [
            190,
            194,
            193,
            192,
        ]
        index = 0
        for itm in tested:
            self.assertEqual(expected[index], itm['id'])
            index += 1

    def test___getitem__(self):
        tested = self.get_object()
        expected = [
            190,
            194,
            193,
            192,
        ]
        for index in range(0, 2, 1):
            self.assertEqual(expected[index], tested[index]['id'])

    def test___bool__(self):
        # with records --> true
        tested = self.get_object()
        self.assertTrue(tested)
        # no record --> false
        tested = UnitTestARecordSet([])
        self.assertFalse(tested)

    def test___len__(self):
        # with records
        tested = self.get_object()
        self.assertEqual(4, len(tested))
        # no record
        tested = UnitTestARecordSet([])
        self.assertEqual(0, len(tested))

    def test___add__(self):
        tested = self.get_object()
        # different class
        other = UnitTestBRecordSet([{'id': 133}])
        with self.assertRaises(Exception):
            result = tested + other

        # same class
        other = UnitTestARecordSet([{'id': 133}])
        result = tested + other
        self.assertEqual(5, len(result))
        expected = [
            190,
            194,
            193,
            192,
            133,
        ]
        index = 0
        for itm in result:
            self.assertEqual(expected[index], itm['id'])
            index += 1

    def test_new_from(self):
        records = [
            {
                'id': 190,
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34560'
                    }]
                }
            },
            {
                'id': 191,
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34561'
                    }]
                }
            },
        ]  # yapf: disable

        tested = UnitTestARecordSet([])
        self.assertEqual(0, len(tested))

        result = tested.new_from(records)
        self.assertIsInstance(result, UnitTestARecordSet)
        self.assertEqual(2, len(result))
        ids = [
            190,
            191,
        ]
        self.assertEqual(ids, [item['id'] for item in result])

    def test_item_to_codes(self):
        tests = [{
            'exp': [],
            'itm': {
                'codes': {
                    'coding': [{
                        'id': 111,
                        'system': 'ICD-10',
                    }]
                },
                'codings': [{
                    'id': 123,
                    'system': 'ICD-10',
                }]
            }
        }, {
            'exp': [{
                'id': 111,
                'system': 'ICD-10',
            }],
            'itm': {
                'code': {
                    'coding': [{
                        'id': 111,
                        'system': 'ICD-10',
                    }]
                },
                'coding': [{
                    'id': 123,
                    'system': 'ICD-10',
                }]
            }
        }, {
            'exp': [{
                'id': 123,
                'system': 'ICD-10',
            }],
            'itm': {
                'cod': {
                    'coding': [{
                        'id': 111,
                        'system': 'ICD-10',
                    }]
                },
                'coding': [{
                    'id': 123,
                    'system': 'ICD-10',
                }]
            }
        }]  # yapf: disable
        for test in tests:
            expected = test['exp']
            result = PatientRecordSet.item_to_codes(test['itm'])
            self.assertEqual(expected, result)

    def test_validated(self):
        # no record with committer --> no record
        tested = self.get_object()
        result = tested.validated()
        self.assertIsInstance(result, UnitTestARecordSet)
        self.assertEqual(0, len(result))

        # record with committer
        tested += UnitTestARecordSet([
            {
                'id': 231,
                'committer': None
            },
            {
                'id': 232,
                'committer': 3,
                'deleted': False
            },
            {
                'id': 233,
                'committer': 3,
                'deleted': True
            },
            {
                'id': 234,
                'committer': 3
            },
            {
                'id': 235,
                'committer': 3,
                'enteredInError': None
            },
            {
                'id': 236,
                'committer': 3,
                'enteredInError': 5
            },
            {
                'id': 237,
                'committer': 3,
                'entered_in_error': None
            },
            {
                'id': 238,
                'committer': 3,
                'entered_in_error': 7
            },
        ])
        result = tested.validated()
        self.assertIsInstance(result, UnitTestARecordSet)
        self.assertEqual(4, len(result))
        self.assertEqual([232, 234, 235, 237], [p['id'] for p in result])

    def test__enumerate_system_list_codes(self):

        class UnitTestReco(ValueSet):
            OID = '1.2.3.4'
            VALUE_SET_NAME = 'Reco for the unit tests'
            CPT = {'abc', '123'}
            SNOMEDCT = {'12345', '23456', '34561', '34562', '34563', '34564'}

        tested = self.get_object()
        expected = {
            'id': [
                190,
                194,
                192,
                193,
            ],
            'coding': [
                {
                    'system': 'http://snomed.info/sct',
                    'code': '34561',
                },
                {
                    'system': 'http://snomed.info/sct',
                    'code': '34562',
                },
                {
                    'system': 'http://snomed.info/sct',
                    'code': '34563',
                },
                {
                    'system': 'http://snomed.info/sct',
                    'code': '34564',
                },
            ]
        }
        count = 0
        for item, code, list_codes in tested._enumerate_system_list_codes(UnitTestReco):
            count = count + 1
            self.assertIn(item['id'], expected['id'])
            self.assertIn(code, expected['coding'])
            self.assertEqual({'12345', '34561', '34563', '34564', '23456', '34562'}, list_codes)
        self.assertEqual(4, count)

    def test_filter(self):
        tested = self.get_object()

        # Test simple filter of ids
        self.assertEqual(1, len(tested.filter(id=193)))
        self.assertEqual(2, len(tested.filter(id__lt=193)))
        self.assertEqual(3, len(tested.filter(id__lte=193)))
        self.assertEqual(2, len(tested.filter(id__gte=193)))
        self.assertEqual(1, len(tested.filter(id__gt=193)))

        # Test nested filter
        self.assertEqual(tested.filter(nested__id=1)[0]['id'], 192)
        self.assertEqual(len(tested.filter(nested__timestamp__gte=arrow.get("2021-01-01"))), 1)

    def test_find(self):

        class UnitTestReco(ValueSet):
            OID = '1.2.3.4'
            VALUE_SET_NAME = 'Reco for the unit tests'
            CPT = {'abc', '123'}
            SNOMEDCT = {'12345', '23456', '34567'}

        # no valid system --> error
        tested = UnitTestNoSystemRecordSet([])
        with self.assertRaises(Exception):
            tested.find(UnitTestReco)

        # valid system
        # - no record with code / coding
        tested = self.get_object()
        result = tested.find(UnitTestReco)
        self.assertIsInstance(result, UnitTestARecordSet)
        self.assertEqual(0, len(result))

        # - record with code / coding + no valid system
        tested += UnitTestARecordSet([{
            'id': 190,
            'code': {
                'coding': [{
                    'id': 101,
                    'system': 'ICD-10',
                    'code': '34561'
                }]
            }
        }])
        result = tested.find(UnitTestReco)
        self.assertIsInstance(result, UnitTestARecordSet)
        self.assertEqual(0, len(result))

        # - record with code / coding + valid system + no valid code
        tested += UnitTestARecordSet([{
            'id': 191,
            'code': {
                'coding': [{
                    'id': 111,
                    'system': 'http://snomed.info/sct',
                    'code': '11111'
                }]
            }
        }])
        result = tested.find(UnitTestReco)
        self.assertIsInstance(result, UnitTestARecordSet)
        self.assertEqual(0, len(result))

        # - record with code / coding + invalid system + valid code
        tested += UnitTestARecordSet([{
            'id': 192,
            'code': {
                'coding': [{
                    'id': 121,
                    'system': 'NOPE',
                    'code': '34567'
                }]
            }
        }])
        result = tested.find(UnitTestReco)
        self.assertIsInstance(result, UnitTestARecordSet)
        self.assertEqual(0, len(result))

        # - record with code / coding + valid system + valid code
        tested += UnitTestARecordSet([{
            'id': 192,
            'code': {
                'coding': [{
                    'id': 121,
                    'system': 'http://snomed.info/sct',
                    'code': '34567'
                }]
            }
        }])
        result = tested.find(UnitTestReco)
        self.assertIsInstance(result, UnitTestARecordSet)
        self.assertEqual(1, len(result))

        # - record with code / coding + valid system + valid code
        tested += UnitTestARecordSet([{
            'id': 193,
            'code': {
                'coding': [{
                    'id': 122,
                    'system': 'http://snomed.info/sct',
                    'code': '34567'
                }]
            }
        }])
        result = tested.find(UnitTestReco)
        self.assertIsInstance(result, UnitTestARecordSet)
        self.assertEqual(2, len(result))
        expected = [192, 193]
        for rec in result:
            exp_id = expected.pop(0)
            self.assertEqual(exp_id, rec['id'])

    def test_find_class(self):

        class UnitTestReco(ValueSet):
            OID = '1.2.3.4'
            VALUE_SET_NAME = 'Reco for the unit tests'
            CPT = {'abc', '123'}
            SNOMEDCT = {r'123\d*', r'[A-Z]{2}\d{2}', '34567'}

        tests = [
            '34567',
            '12345',
            'GO67',
            'GO6745',
        ]
        for code in tests:
            # no valid system --> error
            tested = UnitTestNoSystemRecordSet([])
            with self.assertRaises(Exception):
                tested.find_class(UnitTestReco)

            # valid system
            # - no record with code / coding
            tested = self.get_object()
            result = tested.find_class(UnitTestReco)
            self.assertIsInstance(result, UnitTestARecordSet)
            self.assertEqual(0, len(result))

            # - record with code / coding + no valid system
            tested += UnitTestARecordSet([{
                'id': 190,
                'code': {
                    'coding': [{
                        'id': 101,
                        'system': 'ICD-10',
                        'code': code
                    }]
                }
            }])
            result = tested.find_class(UnitTestReco)
            self.assertIsInstance(result, UnitTestARecordSet)
            self.assertEqual(0, len(result))

            # - record with code / coding + valid system + no valid code
            tested += UnitTestARecordSet([{
                'id': 191,
                'code': {
                    'coding': [{
                        'id': 111,
                        'system': 'http://snomed.info/sct',
                        'code': '11111'
                    }]
                }
            }])
            result = tested.find_class(UnitTestReco)
            self.assertIsInstance(result, UnitTestARecordSet)
            self.assertEqual(0, len(result))

            # - record with code / coding + invalid system + valid code
            tested += UnitTestARecordSet([{
                'id': 192,
                'code': {
                    'coding': [{
                        'id': 121,
                        'system': 'NOPE',
                        'code': code
                    }]
                }
            }])
            result = tested.find_class(UnitTestReco)
            self.assertIsInstance(result, UnitTestARecordSet)
            self.assertEqual(0, len(result))

            # - record with code / coding + valid system + valid code
            tested += UnitTestARecordSet([{
                'id': 192,
                'code': {
                    'coding': [{
                        'id': 121,
                        'system': 'http://snomed.info/sct',
                        'code': code
                    }]
                }
            }])
            result = tested.find_class(UnitTestReco)
            self.assertIsInstance(result, UnitTestARecordSet)
            self.assertEqual(1, len(result))

            # - record with code / coding + valid system + valid code
            tested += UnitTestARecordSet([{
                'id': 193,
                'code': {
                    'coding': [{
                        'id': 122,
                        'system': 'http://snomed.info/sct',
                        'code': code
                    }]
                }
            }])
            result = tested.find_class(UnitTestReco)
            self.assertIsInstance(result, UnitTestARecordSet)
            self.assertEqual(2, len(result))
            expected = [192, 193]
            for rec in result:
                exp_id = expected.pop(0)
                self.assertEqual(exp_id, rec['id'])

    # --- helper ----
    def get_object(self):
        self.assertTrue(issubclass(UnitTestARecordSet, PatientRecordSet))
        records = [
            {
                'id': 190,
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34561'
                    }]
                }
            },
            {
                'id': 194,
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34564'
                    }]
                }
            },
            {
                'id': 193,
                'nested': {
                    'id': 2,
                    'timestamp': '2020-01-01T15:41:14.597558Z'
                },
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34563'
                    }]
                }
            },
            {
                'id': 192,
                'nested': {
                    'id': 1,
                    'timestamp': '2021-08-11T15:41:14.597558Z'
                },
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34562'
                    }]
                }
            },
        ]
        return UnitTestARecordSet(records)


class TestPatientRecordSetEvent(TestCase):

    def test___init__(self):
        tested = self.get_object()
        result = tested.records
        expected = [
            138,
            123,
            456,
        ]
        self.assertEqual(len(expected), len(result))
        for exp, res in zip(expected, result):
            self.assertEqual(exp, res['id'])

        self.assertEqual(expected, [item['id'] for item in result])

    def test_filter_function(self):

        def filter_fn(an_arrow: arrow.Arrow) -> bool:
            return an_arrow.datetime.hour == 13

        # date field empty
        records = [
            {
                'id': 190,
                'fldDate': '',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
            {
                'id': 191,
                'fldDate': '',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
            {
                'id': 192,
                'fldDate': '',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
            {
                'id': 193,
                'fldDate': '',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
        ]
        tested = UnitTestRecordSetEvent(records)
        result = tested.filter_function(filter_fn)
        self.assertEqual(0, len(result))

        # date field present
        tested = self.get_object_with_data()
        result = tested.filter_function(filter_fn)
        self.assertEqual(2, len(result))
        expected = [190, 192]
        for rec in result:
            exp_id = expected.pop(0)
            self.assertEqual(exp_id, rec['id'])

    def test_within(self):
        tested = self.get_object_with_data()
        fmt = 'YYYY-MM-DD HH:mm:ss'
        tests = [
            [
                arrow.get('2018-08-20 13:32:00', fmt),
                arrow.get('2018-08-20 14:33:00', fmt), [192, 193]
            ],
            [arrow.get('2018-08-20 13:32:45', fmt),
             arrow.get('2018-08-20 13:32:45', fmt), [192]],
            [arrow.get('2018-08-20 13:32:00', fmt),
             arrow.get('2018-08-20 13:32:44', fmt), []],
        ]
        for (start, end, expected) in tests:
            tfr = Timeframe(start, end)
            result = tested.within(tfr)
            self.assertEqual(len(expected), len(result))
            for rec in result:
                exp_id = expected.pop(0)
                self.assertEqual(exp_id, rec['id'])

    def test_before(self):
        tested = self.get_object_with_data()

        fmt = 'YYYY-MM-DD HH:mm:ss'
        tests = [
            [arrow.get('2018-08-20 13:33:33', fmt), [190, 191, 192]],
            [arrow.get('2018-08-19 13:32:45', fmt), [190]],
            [arrow.get('2018-08-19 13:32:44', fmt), []],
        ]
        for (end, expected) in tests:
            result = tested.before(end)
            self.assertEqual(len(expected), len(result))
            for rec in result:
                exp_id = expected.pop(0)
                self.assertEqual(exp_id, rec['id'])

    def test_after(self):
        tested = self.get_object_with_data()

        fmt = 'YYYY-MM-DD HH:mm:ss'
        tests = [
            [arrow.get('2018-08-20 12:32:00', fmt), [191, 192, 193]],
            [arrow.get('2018-08-20 13:32:45', fmt), [192, 193]],
            [arrow.get('2018-08-20 14:32:45', fmt), [193]],
            [arrow.get('2018-08-20 14:32:46', fmt), []],
        ]
        for (start, expected) in tests:
            result = tested.after(start)
            self.assertEqual(len(expected), len(result))
            for rec in result:
                exp_id = expected.pop(0)
                self.assertEqual(exp_id, rec['id'])

    def test_within_one_year(self):
        tested = self.get_object_per_year()
        fmt = 'YYYY-MM-DD HH:mm:ss'
        tests = [
            [
                arrow.get('2000-08-20 13:32:00', fmt),
                arrow.get('2019-03-20 13:32:44', fmt), [191, 192]
            ],
            [arrow.get('2000-08-20 13:32:00', fmt),
             arrow.get('2020-03-20 13:32:45', fmt), [193]],
            [arrow.get('2000-08-20 13:32:00', fmt),
             arrow.get('2020-03-20 13:32:46', fmt), []],
        ]
        for (start, end, expected) in tests:
            tfr = Timeframe(start, end)
            result = tested.within_one_year(tfr)
            self.assertEqual(len(expected), len(result))
            for rec in result:
                exp_id = expected.pop(0)
                self.assertEqual(exp_id, rec['id'])

    def test_within_two_years(self):
        tested = self.get_object_per_year()
        fmt = 'YYYY-MM-DD HH:mm:ss'
        tests = [
            [
                arrow.get('2000-08-20 13:32:00', fmt),
                arrow.get('2019-03-20 13:32:44', fmt), [190, 191, 192]
            ],
            [
                arrow.get('2000-08-20 13:32:00', fmt),
                arrow.get('2020-03-20 13:32:44', fmt), [191, 192, 193]
            ],
            [arrow.get('2000-08-20 13:32:00', fmt),
             arrow.get('2021-03-20 13:32:45', fmt), [193]],
            [arrow.get('2000-08-20 13:32:00', fmt),
             arrow.get('2022-03-20 13:32:46', fmt), []],
        ]
        for (start, end, expected) in tests:
            tfr = Timeframe(start, end)
            result = tested.within_two_years(tfr)
            self.assertEqual(len(expected), len(result))
            for rec in result:
                exp_id = expected.pop(0)
                self.assertEqual(exp_id, rec['id'])

    def test_first(self):
        # no record
        tested = UnitTestRecordSetEvent([])
        self.assertIsNone(tested.first())
        # with record
        tested = self.get_object()
        result = tested.first()
        self.assertEqual(138, result['id'])

    def test_last(self):
        # no record
        tested = UnitTestRecordSetEvent([])
        self.assertIsNone(tested.last())
        # with record
        tested = self.get_object()
        result = tested.last()
        self.assertEqual(456, result['id'])

    def test_last_value(self):
        # no record
        tested = UnitTestRecordSetEvent([])
        self.assertIsNone(tested.last_value())
        # with record without casting
        tested = self.get_object()
        result = tested.last_value()
        self.assertEqual('133.55', result)
        # with record with casting
        tested = self.get_object()
        result = tested.last_value(float)
        self.assertEqual(133.55, result)
        # incorrect casting
        result = tested.last_value(int)
        self.assertIsNone(result)

    # --- helpers
    def get_object(self):
        self.assertTrue(issubclass(UnitTestRecordSetEvent, PatientEventRecordSet))
        records = [
            {
                'id': 123,
                'fldDate': '2018-08-19 12:32:45',
                'value': '123AB'
            },
            {
                'id': 456,
                'fldDate': '2018-08-19 12:32:59',
                'value': '133.55'
            },
            {
                'id': 459,
                'fldDate': None
            },
            {
                'id': 138,
                'fldDate': '2018-08-19 10:32:45',
                'value': '143AB'
            },
        ]
        return UnitTestRecordSetEvent(records)

    def get_object_with_data(self):
        self.assertTrue(issubclass(UnitTestRecordSetEvent, PatientEventRecordSet))
        records = [
            {
                'id': 190,
                'fldDate': '2018-08-19 13:32:45',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
            {
                'id': 191,
                'fldDate': '2018-08-20 12:32:45',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
            {
                'id': 192,
                'fldDate': '2018-08-20 13:32:45',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
            {
                'id': 193,
                'fldDate': '2018-08-20 14:32:45',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
        ]
        return UnitTestRecordSetEvent(records)

    def get_object_per_year(self):
        self.assertTrue(issubclass(UnitTestRecordSetEvent, PatientRecordSet))
        records = [
            {
                'id': 190,
                'fldDate': '2018-03-19 13:32:45',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
            {
                'id': 191,
                'fldDate': '2018-07-20 13:32:45',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
            {
                'id': 192,
                'fldDate': '2018-11-20 13:32:45',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
            {
                'id': 193,
                'fldDate': '2019-03-20 13:32:45',
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34567'
                    }]
                }
            },
        ]
        return UnitTestRecordSetEvent(records)


class TestPatientRecordSetPeriod(TestCase):

    def test___init__(self):
        tested = self.get_object()
        result = tested.records
        expected = [
            190,
            194,
            192,
        ]
        self.assertEqual(len(expected), len(result))
        for exp, res in zip(expected, result):
            self.assertEqual(exp, res['id'])

        self.assertEqual(expected, [item['id'] for item in result])

    def test_intersects(self):
        tested = self.get_object()

        timeframe = Timeframe(start=arrow.get('2018-07-15'), end=arrow.get('2018-08-02'))
        result = tested.intersects(timeframe, still_active=False)
        expected = []
        self.assertEqual(expected, [item['id'] for item in result])

        timeframe = Timeframe(start=arrow.get('2018-07-15'), end=arrow.get('2018-08-20'))
        result = tested.intersects(timeframe, still_active=False)
        expected = [
            190,
            194,
        ]
        self.assertEqual(expected, [item['id'] for item in result])

        timeframe = Timeframe(start=arrow.get('2018-07-15'), end=arrow.get('2018-12-20'))
        result = tested.intersects(timeframe, still_active=False)
        expected = [
            190,
            194,
            192,
        ]
        self.assertEqual(expected, [item['id'] for item in result])

        timeframe = Timeframe(start=arrow.get('2018-10-15'), end=arrow.get('2018-11-20'))
        result = tested.intersects(timeframe, still_active=False)
        expected = [
            190,
            194,
        ]
        self.assertEqual(expected, [item['id'] for item in result])

        timeframe = Timeframe(start=arrow.get('2018-12-15'), end=arrow.get('2018-12-20'))
        result = tested.intersects(timeframe, still_active=False)
        expected = [
            190,
            192,
        ]
        self.assertEqual(expected, [item['id'] for item in result])

        timeframe = Timeframe(start=arrow.get('2018-09-05'), end=arrow.get('2018-11-15'))
        result = tested.intersects(timeframe, still_active=False)
        expected = [
            190,
            194,
        ]
        self.assertEqual(expected, [item['id'] for item in result])
        result = tested.intersects(timeframe, still_active=True)
        expected = [
            190,
        ]
        self.assertEqual(expected, [item['id'] for item in result])

        timeframe = Timeframe(start=arrow.get('2018-09-05'), end=arrow.get('2018-11-05'))
        result = tested.intersects(timeframe, still_active=False)
        expected = [
            190,
            194,
        ]
        self.assertEqual(expected, [item['id'] for item in result])
        result = tested.intersects(timeframe, still_active=True)
        expected = [
            190,
            194,
        ]
        self.assertEqual(expected, [item['id'] for item in result])

    def test_starts_before(self):
        tested = self.get_object()

        result = tested.starts_before(arrow.get('2018-07-15'))
        expected = []
        self.assertEqual(expected, [item['id'] for item in result])

        result = tested.starts_before(arrow.get('2018-09-15'))
        expected = [
            190,
        ]
        self.assertEqual(expected, [item['id'] for item in result])

        result = tested.starts_before(arrow.get('2018-12-15'))
        expected = [
            190,
            192,
        ]
        self.assertEqual(expected, [item['id'] for item in result])

    # --- helper ----
    def get_object(self):
        self.assertTrue(issubclass(UnitTestRecordSetPeriod, PatientPeriodRecordSet))
        records = [
            {
                'id': 190,
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34561'
                    }]
                },
                'periods': [
                    {
                        'from': '2018-08-15',
                        'to': '2018-08-25'
                    },
                    {
                        'from': '2018-09-05',
                        'to': None
                    },
                ]
            },
            {
                'id': 194,
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34564'
                    }]
                },
                'periods': [
                    {
                        'from': '2018-08-15',
                        'to': '2018-08-25'
                    },
                    {
                        'from': '2018-09-10',
                        'to': '2018-11-10'
                    },
                ]
            },
            {
                'id': 193,
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34563'
                    }]
                },
                'periods': []
            },
            {
                'id': 192,
                'code': {
                    'coding': [{
                        'system': 'http://snomed.info/sct',
                        'code': '34562'
                    }]
                },
                'periods': [{
                    'from': '2018-11-30',
                    'to': None
                }]
            },
        ]
        return UnitTestRecordSetPeriod(records)


class TestBillingLineItemRecordSet(TestCase):

    def test_constants(self):
        tested = BillingLineItemRecordSet([])
        self.assertIsInstance(tested, BillingLineItemRecordSet)
        self.assertEqual(['cpt', 'hcpcs', 'snomedct'], tested.VALID_SYSTEMS)
        self.assertEqual('datetimeOfService', tested.DATE_FIELD)
        self.assertEqual('billing_line_items', tested.PATIENT_FIELD)
        self.assertEqual('patient__key', tested.API_KEY_FIELD)
        self.assertEqual('billingLineItems', tested.API_UPDATE_FIELD)


class TestConditionRecordSet(TestCase):

    def test_constants(self):
        tested = ConditionRecordSet([])
        self.assertIsInstance(tested, PatientPeriodRecordSet)
        self.assertEqual(['icd10cm', 'icd10pcs', 'snomedct'], tested.VALID_SYSTEMS)
        self.assertEqual('conditions', tested.PATIENT_FIELD)
        self.assertEqual('patient__key', tested.API_KEY_FIELD)
        self.assertEqual('conditions', tested.API_UPDATE_FIELD)

    def test_item_to_codes(self):
        tested = ConditionRecordSet
        tests = [
            ('active', True),
            ('remission', False),
            ('relapse', False),
            ('resolved', True),
            ('investigative', False),
        ]
        for status, expected in tests:
            item = {
                'clinicalStatus': status,
                'coding': [{
                    'system': 'ICD-10',
                    'code': 'I420'
                }],
                'periods': [{
                    'from': '2018-08-23',
                    'to': None
                }],
            }
            result = tested.item_to_codes(item)
            if expected:
                self.assertEqual(item['coding'], result)
            else:
                self.assertEqual([], result)


class TestImmunizationRecordSet(TestCase):

    def test_constants(self):
        tested = ImmunizationRecordSet([])
        self.assertIsInstance(tested, PatientEventRecordSet)
        self.assertEqual(['cvx'], tested.VALID_SYSTEMS)
        self.assertEqual('dateOrdered', tested.DATE_FIELD)
        self.assertEqual('immunizations', tested.PATIENT_FIELD)
        self.assertEqual('patient__key', tested.API_KEY_FIELD)
        self.assertEqual('immunizations', tested.API_UPDATE_FIELD)

    def test_item_to_codes(self):
        item = {
            'codes': {
                'value': '124AB'
            },
            'code': {
                'value': '456AB'
            },
            'coding': {
                'value': '123AB'
            },
            'codings': {
                'value': '789AB'
            },
        }
        self.assertEqual({'value': '123AB'}, ImmunizationRecordSet.item_to_codes(item))


class TestVitalSignRecordSet(TestCase):

    def test_constants(self):
        tested = VitalSignRecordSet([])
        self.assertIsInstance(tested, PatientEventRecordSet)
        self.assertEqual(['loinc'], tested.VALID_SYSTEMS)
        self.assertEqual('dateRecorded', tested.DATE_FIELD)
        self.assertEqual('vital_signs', tested.PATIENT_FIELD)
        self.assertEqual('reading__patient__key', tested.API_KEY_FIELD)
        self.assertEqual('vitalSigns', tested.API_UPDATE_FIELD)

    def test_item_to_codes(self):
        item = {
            'loinxNum': '29463-4',
            'loincNum': '29463-5',
            'loincNux': '29463-6',
        }
        self.assertEqual([{
            'code': '29463-5',
            'system': 'http://loinc.org'
        }], VitalSignRecordSet.item_to_codes(item))


class TestLabReportRecordSet(TestCase):

    def test_constants(self):
        tested = LabReportRecordSet([])
        self.assertIsInstance(tested, PatientEventRecordSet)
        self.assertEqual(['loinc'], tested.VALID_SYSTEMS)
        self.assertEqual('originalDate', tested.DATE_FIELD)
        self.assertEqual('lab_reports', tested.PATIENT_FIELD)
        self.assertEqual('patient__key', tested.API_KEY_FIELD)
        self.assertEqual('labReports', tested.API_UPDATE_FIELD)

    def test_item_to_codes(self):
        item = {
            'loincCodex': [{
                'id': 21,
                'code': '1798-',
                'name': 'Amylasey',
            }],
            'loincCodes': [
                {
                    'id': 38,
                    'code': '1798-8',
                    'name': 'Amylase',
                },
                {
                    'id': 37,
                    'code': '1798-3',
                    'name': 'Amylasx',
                },
                {
                    'id': 39,
                    'code': '1798-',
                    'name': 'Amylasey',
                },
            ],
            'loincCodey': [{
                'id': 32,
                'code': '1798-8',
                'name': 'Amylase',
            }],
        }
        expected = [{
            'system': 'http://loinc.org',
            'id': 38,
            'code': '1798-8',
            'name': 'Amylase',
        }, {
            'system': 'http://loinc.org',
            'id': 37,
            'code': '1798-3',
            'name': 'Amylasx',
        }, {
            'system': 'http://loinc.org',
            'id': 39,
            'code': '1798-',
            'name': 'Amylasey',
        }]
        self.assertEqual(expected, LabReportRecordSet.item_to_codes(item))

    def test_new_from(self):
        records = [{
            'test': 123,
            'values': [
                {
                    'originalDate': '2018-05-29',
                    'id': 911,
                    'loincCodes': [{
                        'id': 38,
                        'created': '2018-05-29T02:38:22.011169Z',
                        'code': '1798-8',
                    }]
                },
                {
                    'originalDate': '2018-05-29',
                    'id': 912,
                    'loincCodes': [{
                        'id': 47,
                        'created': '2018-05-30T02:38:22.011169Z',
                        'code': '1798-8',
                    }]
                },
            ],
            'originalDate': '2018-05-29'
        }]
        tested = LabReportRecordSet([])
        self.assertEqual(0, len(tested.records))

        result = tested.new_from(records)
        self.assertIsInstance(result, LabReportRecordSet)
        self.assertNotEqual(tested, result)

        self.assertEqual(2, len(result))
        expected = [{
            'originalDate': '2018-05-29',
            'id': 911,
            'loincCodes': [{
                'id': 38,
                'created': '2018-05-29T02:38:22.011169Z',
                'code': '1798-8'
            }]
        }, {
            'originalDate': '2018-05-29',
            'id': 912,
            'loincCodes': [{
                'id': 47,
                'created': '2018-05-30T02:38:22.011169Z',
                'code': '1798-8'
            }]
        }]  # yapf: disable
        for exp_item, res_item in zip(expected, result):
            self.assertEqual(exp_item['id'], res_item['id'])


class TestInstructionRecordSet(TestCase):

    def test_constants(self):
        tested = InstructionRecordSet([])
        self.assertIsInstance(tested, PatientEventRecordSet)
        self.assertEqual(['icd10cm', 'icd10pcs', 'snomedct', 'cpt', 'hcpcs'], tested.VALID_SYSTEMS)
        self.assertEqual('noteTimestamp', tested.DATE_FIELD)
        self.assertEqual('instructions', tested.PATIENT_FIELD)
        self.assertEqual('patient__key', tested.API_KEY_FIELD)
        self.assertEqual('instructions', tested.API_UPDATE_FIELD)


class TestMedicationRecordSet(TestCase):

    def test_constants(self):
        tested = MedicationRecordSet([])
        self.assertIsInstance(tested, PatientPeriodRecordSet)
        self.assertEqual([
            'rxnorm',
            'fdb',
        ], tested.VALID_SYSTEMS)
        self.assertEqual('medications', tested.PATIENT_FIELD)
        self.assertEqual('patient__key', tested.API_KEY_FIELD)
        self.assertEqual('medications', tested.API_UPDATE_FIELD)

    def test_item_to_codes(self):
        item = {
            'codinx': [{
                'system': '',
                'code': 'cde113',
                'display': 'The A'
            }],
            'coding': [
                {
                    'system': '',
                    'code': 'cde123',
                    'display': 'The A'
                },
                {
                    'system': '',
                    'code': 'cde124',
                    'display': 'The B'
                },
                {
                    'system': '',
                    'code': 'cde125',
                    'display': 'The C'
                },
            ],
            'codiny': [{
                'system': '',
                'code': 'cde133',
                'display': 'The A'
            }],
        }
        expected = [
            {
                'system': '',
                'code': 'cde123',
                'display': 'The A'
            },
            {
                'system': '',
                'code': 'cde124',
                'display': 'The B'
            },
            {
                'system': '',
                'code': 'cde125',
                'display': 'The C'
            },
        ]
        self.assertEqual(expected, MedicationRecordSet.item_to_codes(item))


class TestPrescriptionRecordSet(TestCase):

    def test_constants(self):
        tested = PrescriptionRecordSet([])
        self.assertIsInstance(tested, PatientRecordSet)
        self.assertEqual(['rxnorm', 'fdb', 'ndc'], tested.VALID_SYSTEMS)
        self.assertEqual('prescriptions', tested.PATIENT_FIELD)
        self.assertEqual('prescriptions', tested.API_UPDATE_FIELD)

    def test_item_to_codes(self):
        item = {
            'codinx': [{
                'system': '',
                'code': 'cde113',
                'display': 'The A'
            }],
            'coding': [
                {
                    'system': '',
                    'code': 'cde123',
                    'display': 'The A'
                },
                {
                    'system': '',
                    'code': 'cde124',
                    'display': 'The B'
                },
                {
                    'system': '',
                    'code': 'cde125',
                    'display': 'The C'
                },
            ],
            'codiny': [{
                'system': '',
                'code': 'cde133',
                'display': 'The A'
            }],
        }
        expected = [
            {
                'system': '',
                'code': 'cde123',
                'display': 'The A'
            },
            {
                'system': '',
                'code': 'cde124',
                'display': 'The B'
            },
            {
                'system': '',
                'code': 'cde125',
                'display': 'The C'
            },
        ]
        self.assertEqual(expected, PrescriptionRecordSet.item_to_codes(item))


class TestInterviewRecordSet(TestCase):

    def test_constants(self):
        tested = InterviewRecordSet([])
        self.assertIsInstance(tested, PatientEventRecordSet)
        self.assertEqual([
            'loinc',
            'snomedct',
            'rxnorm',
            'canvas',
            'internal',
        ], tested.VALID_SYSTEMS)
        self.assertEqual('noteTimestamp', tested.DATE_FIELD)
        self.assertEqual('interviews', tested.PATIENT_FIELD)
        self.assertEqual('patient__key', tested.API_KEY_FIELD)
        self.assertEqual('interviews', tested.API_UPDATE_FIELD)

    def test_item_to_codes(self):
        # flake8: noqa
        item = {
            'results': [
                {'codeSystem': 'http://loina.org', 'code': '76504-0', 'id': 127, },
                {'codeSystem': 'http://loinc.org', 'code': '76504-1', 'id': 128, },
                {'codeSystem': 'http://loinc.org', 'code': '', 'id': 129, },
                {'codeSystem': '', 'code': '76504-3', 'id': 130},
            ],
            'questionnaires': [
                {'id': 9, 'code': '12771-1', 'codeSystem': 'http://loinx.org', },
                {'id': 3, 'code': '12771-3', 'codeSystem': 'http://loinz.org', },
            ],
            'questions': [
                {'id': 129, 'code': '76501-1', 'codeSystem': 'http://loinx.org', 'questionResponseId': 33, },
                {'id': 125, 'code': '76501-2', 'codeSystem': 'http://loiny.org', 'questionResponseId': 34, },
                {'id': 123, 'code': '76501-3', 'codeSystem': 'http://loinz.org', 'questionResponseId': 35, },
            ],
            'responses': [
                {'id': 439, 'code': '96501-1', 'codeSystem': 'http://loinx.org', 'questionResponseId': 33, },
                {'id': 435, 'code': '96501-2', 'codeSystem': 'http://loiny.org', 'questionResponseId': 34, },
                {'id': 433, 'code': '96501-3', 'codeSystem': 'http://loinz.org', 'questionResponseId': 35, },
                {'id': 437, 'code': '96501-4', 'codeSystem': 'http://loinz.org', 'questionResponseId': 35, },
            ],
        }  # yapf: disable
        expected = [
            {'code': '76504-0', 'system': 'http://loina.org', 'question_id': None, },
            {'code': '76504-1', 'system': 'http://loinc.org', 'question_id': None, },
            {'code': '12771-1', 'system': 'http://loinx.org', 'question_id': None, },
            {'code': '12771-3', 'system': 'http://loinz.org', 'question_id': None, },
            {'code': '76501-1', 'system': 'http://loinx.org', 'question_id': 33, },
            {'code': '76501-2', 'system': 'http://loiny.org', 'question_id': 34, },
            {'code': '76501-3', 'system': 'http://loinz.org', 'question_id': 35, },
            {'code': '96501-1', 'system': 'http://loinx.org', 'question_id': 33, },
            {'code': '96501-2', 'system': 'http://loiny.org', 'question_id': 34, },
            {'code': '96501-3', 'system': 'http://loinz.org', 'question_id': 35, },
            {'code': '96501-4', 'system': 'http://loinz.org', 'question_id': 35, },
        ]  # yapf: disable
        self.assertEqual(expected, InterviewRecordSet.item_to_codes(item))

    def test_find_question_response(self):

        class UnitTestQuestion(ValueSet):
            OID = '1.2.3.4'
            VALUE_SET_NAME = 'Question for the unit tests'
            CPT = {'abc', '123'}
            SNOMEDCT = {'234', '456', '675'}

        class UnitTestResponse(ValueSet):
            OID = '1.2.3.5'
            VALUE_SET_NAME = 'Response for the unit tests'
            LOINC = {'xyz', '987'}
            SNOMEDCT = {'543', '654', '765'}

        # codes are in the question and the response
        item = [{
            'noteTimestamp': '2019-03-15 11:54:00+00:00',
            'results': [],
            'questionnaires': [],
            'questions': [
                {'id': 129, 'code': '456', 'codeSystem': 'http://snomed.info/sct', 'questionResponseId': 33, },
            ],
            'responses': [
                {'id': 439, 'code': 'xyz', 'codeSystem': 'http://loinc.org', 'questionResponseId': 33, },
            ],
        }]  # yapf: disable
        tested = InterviewRecordSet(item)
        result = tested.find_question_response(UnitTestQuestion, UnitTestResponse)
        self.assertIsInstance(result, InterviewRecordSet)
        self.assertEqual(1, len(result.records))

        # codes are in different question and response
        item = [{
            'noteTimestamp': '2019-03-15 11:54:00+00:00',
            'results': [],
            'questionnaires': [],
            'questions': [
                {'id': 129, 'code': '456', 'codeSystem': 'http://snomed.info/sct', 'questionResponseId': 33, },
            ],
            'responses': [
                {'id': 439, 'code': 'xyz', 'codeSystem': 'http://loinc.org', 'questionResponseId': 30, },
            ],
        }]  # yapf: disable
        tested = InterviewRecordSet(item)
        result = tested.find_question_response(UnitTestQuestion, UnitTestResponse)
        self.assertIsInstance(result, InterviewRecordSet)
        self.assertEqual(0, len(result.records))


class TestReferralReportRecordSet(TestCase):

    def test_constants(self):
        tested = ReferralReportRecordSet([])
        self.assertIsInstance(tested, PatientEventRecordSet)
        self.assertEqual(['snomedct', 'loinc', 'cpt'], tested.VALID_SYSTEMS)
        self.assertEqual('originalDate', tested.DATE_FIELD)
        self.assertEqual('referral_reports', tested.PATIENT_FIELD)
        self.assertEqual('patient__key', tested.API_KEY_FIELD)
        self.assertEqual('referralReports', tested.API_UPDATE_FIELD)

    def test_item_to_codes(self):
        item = {
            'codings': [{
                'id': 5,
                'system': 'http://www.ama-assn.org/go/cpt',
                'version': '',
                'code': '37231002',
                'value': 'Present',
                'display': 'Macular Edema',
            }, {
                'id': 4,
                'system': 'http://snomed.info/sct',
                'version': '',
                'code': '430801000124103',
                'value': 'Proliferative',
                'display': 'Retinopathy',
            }, {
                'id': 6,
                'system': 'BI-RADS assessment',
                'version': '',
                'code': '24606-6',
                'value': 'benign findings',
                'display': 'BI-RADS assessment',
            }],
        }  # yapf: disable
        expected = [
            {
                'code': '37231002',
                'system': 'http://www.ama-assn.org/go/cpt',
            },
            {
                'code': '430801000124103',
                'system': 'http://snomed.info/sct',
            },
            {
                'code': '24606-6',
                'system': 'BI-RADS assessment',
            },
        ]
        result = ReferralReportRecordSet.item_to_codes(item)
        self.assertEqual(len(expected), len(result))
        for item in result:
            item_present = False
            for option in expected:
                if option['code'] == item['code'] and option['system'] == item['system']:
                    item_present = True
            self.assertTrue(item_present)


class TestImagingReportRecordSet(TestCase):

    def test_constants(self):
        tested = ImagingReportRecordSet([])
        self.assertIsInstance(tested, PatientEventRecordSet)
        self.assertEqual(['snomedct', 'loinc', 'cpt'], tested.VALID_SYSTEMS)
        self.assertEqual('originalDate', tested.DATE_FIELD)
        self.assertEqual('imaging_reports', tested.PATIENT_FIELD)
        self.assertEqual('patient__key', tested.API_KEY_FIELD)
        self.assertEqual('imagingReports', tested.API_UPDATE_FIELD)

    def test_item_to_codes(self):
        item = {
            'codings': [{
                'id': 21,
                'system': 'http://loinc.org',
                'version': '',
                'code': '408573005',
                'value': 'Normal',
                'display': 'Status'
            }, {
                'id': 20,
                'system': 'http://www.ama-assn.org/go/cpt',
                'version': '',
                'code': '282290005',
                'value': '5',
                'display': 'Interpretation'
            }, {
                'id': 4,
                'system': 'http://snomed.info/sct',
                'version': '',
                'code': '430801000124103',
                'value': 'Proliferative',
                'display': 'Retinopathy',
            }, {
                'id': 19,
                'system': 'http://www.ama-assn.org/go/cpt',
                'version': '',
                'code': '281296001',
                'value': '11',
                'display': 'Comment'
            }],
        }  # yapf: disable
        expected = [
            {
                'code': '408573005',
                'system': 'http://loinc.org'
            },
            {
                'code': '282290005',
                'system': 'http://www.ama-assn.org/go/cpt'
            },
            {
                'code': '430801000124103',
                'system': 'http://snomed.info/sct',
            },
            {
                'code': '281296001',
                'system': 'http://www.ama-assn.org/go/cpt'
            },
        ]
        self.assertEqual(expected, ImagingReportRecordSet.item_to_codes(item))

class TestConsentRecordSet(TestCase):

    def test_constants(self):
        tested = ConsentRecordSet([])
        self.assertIsInstance(tested, PatientRecordSet)
        self.assertEqual(['canvas', 'internal'], tested.VALID_SYSTEMS)
        self.assertEqual('consents', tested.PATIENT_FIELD)
        self.assertEqual('consents', tested.API_UPDATE_FIELD)

class TestGroupRecordSet(TestCase):

    def test_constants(self):
        tested = GroupRecordSet([])
        self.assertIsInstance(tested, PatientRecordSet)
        self.assertEqual('groups', tested.PATIENT_FIELD)
        self.assertEqual('groups', tested.API_UPDATE_FIELD)

class TestTaskRecordSet(TestCase):

    def test_constants(self):
        tested = TaskRecordSet([])
        self.assertIsInstance(tested, PatientRecordSet)
        self.assertEqual(['canvas', 'internal'], tested.VALID_SYSTEMS)
        self.assertEqual('tasks', tested.PATIENT_FIELD)
        self.assertEqual('tasks', tested.API_UPDATE_FIELD)

class TestProcedureRecordSet(TestCase):

    def test_constants(self):
        tested = ProcedureRecordSet([])
        self.assertIsInstance(tested, PatientRecordSet)
        self.assertEqual(['cpt', 'snomedct'], tested.VALID_SYSTEMS)
        self.assertEqual('procedures', tested.PATIENT_FIELD)
        self.assertEqual('procedures', tested.API_UPDATE_FIELD)

    def test_item_to_codes(self):
        item = {
            'codinx': [{
                'system': '',
                'code': 'cde113',
                'display': 'The A'
            }],
            'coding': [
                {
                    'system': '',
                    'code': 'cde123',
                    'display': 'The A'
                },
                {
                    'system': '',
                    'code': 'cde124',
                    'display': 'The B'
                },
                {
                    'system': '',
                    'code': 'cde125',
                    'display': 'The C'
                },
            ],
            'codiny': [{
                'system': '',
                'code': 'cde133',
                'display': 'The A'
            }],
        }
        expected = [
            {
                'system': '',
                'code': 'cde123',
                'display': 'The A'
            },
            {
                'system': '',
                'code': 'cde124',
                'display': 'The B'
            },
            {
                'system': '',
                'code': 'cde125',
                'display': 'The C'
            },
        ]
        self.assertEqual(expected, PrescriptionRecordSet.item_to_codes(item))


class TestProtocolOverrideRecordSet(TestCase):

    def test_constants(self):
        tested = ProtocolOverrideRecordSet([])
        self.assertIsInstance(tested, PatientRecordSet)
        self.assertEqual(['snomedct'], tested.VALID_SYSTEMS)
        self.assertEqual('protocol_overrides', tested.PATIENT_FIELD)
        self.assertEqual('patient__key', tested.API_KEY_FIELD)
        self.assertEqual('protocolOverrides', tested.API_UPDATE_FIELD)

    def test_is_switched_off(self):
        tested = ProtocolOverrideRecordSet([{
            'protocolKey': 'TEST001v2',
            'adjustment': {
                'reference': '2018-10-01T00:00:00Z',
                'cycleDays': 19
            },
            'snooze': {
                'reference': '2018-10-01T00:00:00Z',
                'snoozedDays': 22,
                'reasonCode': '413311005',
                'reasonText': 'PATIENT_REFUSED',
                'reasonCodeSystem': 'snomedct'
            },
            'modified': '2018-10-02T21:50:09.661490Z'
        }])  # yapf: disable
        self.assertFalse(tested.all_switched_off)
        tested = ProtocolOverrideRecordSet([{
            'protocolKey': '*',
            'adjustment': {
                'reference': '2018-10-01T00:00:00Z',
                'cycleDays': 19
            },
            'snooze': {
                'reference': '2018-10-01T00:00:00Z',
                'snoozedDays': 22,
                'reasonCode': '413311005',
                'reasonText': 'PATIENT_REFUSED',
                'reasonCodeSystem': 'snomedct'
            },
            'modified': '2018-10-02T21:50:09.661490Z'
        }])  # yapf: disable
        self.assertTrue(tested.all_switched_off)

    def test_is_snoozed(self):
        tested = ProtocolOverrideRecordSet([
            {
                'protocolKey': 'TEST006v2',
                'adjustment': None,
                'snooze': {
                    'reference': '2018-09-15T00:00:00Z',
                    'snoozedDays': 12,
                    'reasonCode': '413311005',
                    'reasonText': 'PATIENT_REFUSED',
                    'reasonCodeSystem': 'snomedct'
                },
                'modified': '2018-09-15T21:50:09.661490Z'
            },
            {
                'protocolKey': 'TEST001v2',
                'adjustment': {
                    'reference': '2018-10-01T00:00:00Z',
                    'cycleDays': 19
                },
                'snooze': {
                    'reference': arrow.now(),
                    'snoozedDays': 22,
                    'reasonCode': '413311005',
                    'reasonText': 'PATIENT_REFUSED',
                    'reasonCodeSystem': 'snomedct'
                },
                'modified': '2018-10-02T21:50:09.661490Z'
            },
            {
                'protocolKey': 'TEST003v2',
                'adjustment': None,
                'snooze': {
                    'reference': '2018-09-15T00:00:00Z',
                    'snoozedDays': 12,
                    'reasonCode': '413311005',
                    'reasonText': 'PATIENT_REFUSED',
                    'reasonCodeSystem': 'snomedct'
                },
                'modified': '2018-09-15T21:50:09.661490Z'
            },
            {
                'protocolKey': 'TEST004v1',
                'adjustment': {
                    'reference': '2018-08-15T00:00:00Z',
                    'cycleDays': 60
                },
                'snooze': None,
                'modified': '2018-08-23T21:50:09.661490Z'
            }])  # yapf: disable
        # protocol is not overridden --> false
        self.assertFalse(tested.is_snoozed(['TEST005v1']))
        # snooze is not defined --> false
        self.assertFalse(tested.is_snoozed(['TEST004v1']))
        # snooze defined but past --> false
        self.assertFalse(tested.is_snoozed(['TEST003v2']))
        self.assertFalse(tested.is_snoozed([
            'TEST003v2',
            'TEST004v1',
            'TEST005v1',
        ]))
        # snooze defined and not past yet --> true
        self.assertTrue(tested.is_snoozed(['TEST001v2']))
        self.assertTrue(tested.is_snoozed([
            'TEST003v2',
            'TEST001v2',
            'TEST005v1',
        ]))

    def test_item_to_codes(self):
        item = {
            'protocolKey': 'TEST001v2',
            'adjustment': {
                'reference': '2018-10-01T00:00:00Z',
                'cycleDays': 19
            },
            'snooze': {
                'reference': arrow.now(),
                'snoozedDays': 22,
                'reasonCode': '413311005',
                'reasonText': 'PATIENT_REFUSED',
                'reasonCodeSystem': 'snomedct'
            },
            'modified': '2018-10-02T21:50:09.661490Z'
        }  # yapf: disable
        expected = [
            {
                'code': '413311005',
                'system': 'snomedct'
            },
        ]
        self.assertEqual(expected, ProtocolOverrideRecordSet.item_to_codes(item))


# --- helpers ---
class UnitTestRecordSetEvent(PatientEventRecordSet):
    VALID_SYSTEMS = ['icd10cm', 'icd10pcs', 'snomedct']
    DATE_FIELD = 'fldDate'
    PATIENT_FIELD = 'fldPatient'
    API_UPDATE_FIELD = 'updateA'


class UnitTestRecordSetPeriod(PatientPeriodRecordSet):
    VALID_SYSTEMS = ['icd10cm', 'snomedct']
    DATE_FIELD = 'fldDate'
    PATIENT_FIELD = 'fldPatient'
    API_KEY_FIELD = 'another__key'
    API_UPDATE_FIELD = 'updateB'


class UnitTestARecordSet(PatientRecordSet):
    VALID_SYSTEMS = ['icd10cm', 'snomedct']
    PATIENT_FIELD = 'fldPatient'
    API_UPDATE_FIELD = 'updateD'


class UnitTestBRecordSet(PatientRecordSet):
    VALID_SYSTEMS = ['icd10cm']
    PATIENT_FIELD = 'fldPatient'
    API_KEY_FIELD = 'another__key'
    API_UPDATE_FIELD = 'updateE'


class UnitTestNoSystemRecordSet(PatientRecordSet):
    VALID_SYSTEMS = []
    PATIENT_FIELD = 'fldPatient'
    API_UPDATE_FIELD = 'updateC'



# type: ignore
from typing import Dict, List, Optional, cast

import arrow

from cached_property import cached_property

from canvas_workflow_kit import events, settings
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    CodingStruct,
    ExternallyAwareClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import StructuredAssessmentRecommendation
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.value_set.customer_value_sets.scan import (
    QuestionnairePVQ,
    StructuredAssessmentQuestionnaireUrinarySymptoms
)


class CustomerScanUrinarySymptomsAssessment(ExternallyAwareClinicalQualityMeasure,
                                            ClinicalQualityMeasure):
    """
    If a patient has urinary leakage, administer a Urinary Symptoms Structured Assessment.
    """

    class Meta:
        title = 'SCAN: Urinary Symptoms Assessment'
        version = '2020-10-13v1'
        description = ''
        information = 'https://docs.google.com/document/d/1NW4b3OT_yPeZy_oIKt_DCCeQ0egDaWFE/edit'

        identifiers = ['CustomerScanUrinarySymptomsAssessmentV1']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        authors = [
            'SCAN',
        ]
        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_INTERVIEW,
        ]

        funding_source = 'SCAN'

        references = ['Written by SCAN']

    @classmethod
    def enabled(cls) -> bool:
        return cast(bool, settings.IS_HEALTHCHEC)

    def in_initial_population(self) -> bool:
        return True

    def interview_lookback_timeframe(self) -> Timeframe:
        return Timeframe(start=self.timeframe.end.shift(days=-30), end=self.timeframe.end)

    @cached_property
    def recent_pvqs(self) -> List[Dict]:
        interviews = self.patient.interviews.find(QuestionnairePVQ).within(
            self.interview_lookback_timeframe())

        for interview in interviews:
            interview['groupedResponses'] = self.group_question_responses(interview)

        return interviews

    @cached_property
    def latest_urine_leakage_response(self) -> Optional[Dict]:
        urine_leakage_question_coding = CodingStruct(code='123978lijsdf386', system='INTERNAL')

        return self.latest_response_to_question(question_coding=urine_leakage_question_coding)

    def latest_urine_leakage_response_date(self) -> Optional[arrow.Arrow]:
        if self.latest_urine_leakage_response:
            return arrow.get(self.latest_urine_leakage_response.get('note_timestamp'))

    def latest_response_to_question(self, question_coding: CodingStruct) -> Optional[Dict]:
        for interview in sorted(
                self.recent_pvqs, key=lambda i: arrow.get(i['noteTimestamp']), reverse=True):
            question_responses = interview.get('groupedResponses', {}).get(
                question_coding.code, [])
            if question_responses:
                return {
                    'note_timestamp': interview['noteTimestamp'],
                    'question_coding': question_coding,
                    'responses': question_responses,
                }

    def last_interview(self, questionnaire_valueset, timeframe=None) -> Optional[Dict]:
        interview = self.patient.interviews.find(questionnaire_valueset).within(
            timeframe or self.timeframe).last()

        if not interview:
            return None

        return interview

    @cached_property
    def last_urinary_symptoms_assessment(self) -> Optional[Dict]:
        return self.last_interview(
            questionnaire_valueset=StructuredAssessmentQuestionnaireUrinarySymptoms,
            timeframe=self.interview_lookback_timeframe())

    def last_urinary_symptoms_assessment_date(self) -> Optional[arrow.Arrow]:
        if self.last_urinary_symptoms_assessment:
            return arrow.get(self.last_urinary_symptoms_assessment['noteTimestamp'])
        return None

    def urinary_symptoms_assessed_on_or_after(self, reference_date: arrow.Arrow) -> bool:
        # It can't have been assessed after a date that doesn't exist!
        if not reference_date:
            return False

        # If it hasn't been assessed, it certainly wasn't assessed after the
        # reference date.
        if not self.last_urinary_symptoms_assessment_date():
            return False

        return self.last_urinary_symptoms_assessment_date() >= reference_date

    def in_denominator(self) -> bool:
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients who were in hospice care during the measurement year

        Exceptions: None
        """
        if not self.in_initial_population():
            return False

        if self.patient.hospice_within(self.timeframe):
            return False

        return True

    def in_numerator(self) -> bool:
        """
        Patient is in the numerator if they have not indicated (via the PVQ) that they have
        experienced urine leakage OR if they have indicated experiencing urine leakage, but
        have since had their urinary symptoms assessed.
        """

        urine_leakage_affirmative_answer_coding = CodingStruct(
            code='123978lijsdf387', system='INTERNAL')

        urine_leakage_reported = False

        if self.latest_urine_leakage_response:
            urine_leakage_reported = urine_leakage_affirmative_answer_coding.code in self.latest_urine_leakage_response[  # noqa: E501
                'responses']

        no_urine_leakage_reported = not urine_leakage_reported

        urinary_symptoms_assessed = self.urinary_symptoms_assessed_on_or_after(
            self.latest_urine_leakage_response_date())

        return no_urine_leakage_reported or (urine_leakage_reported and urinary_symptoms_assessed)

    def craft_satisfied_result(self):
        """
        Satisfied if in numerator. (See in_numerator for inclusion details)
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_SATISFIED

        urinary_symptoms_assessed = self.urinary_symptoms_assessed_on_or_after(
            self.latest_urine_leakage_response_date())
        if urinary_symptoms_assessed:
            result.add_narrative(
                f'{self.patient.first_name} reported urinary leakage, but their symptoms were '
                f'assessed {self.display_date(self.last_urinary_symptoms_assessment_date())}.')
        else:
            result.add_narrative(f'{self.patient.first_name} has not reported urinary leakage '
                                 'within the past 30 days.')

        return result

    def craft_unsatisfied_result(self):
        """
        Unsatisfied if not in numerator. Recommend the Urinary Symptoms Structured Assessment.
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_DUE

        result.add_narrative(f'{self.patient.first_name} reported urinary leakage '
                             f'{self.display_date(self.latest_urine_leakage_response_date())}.')

        result.add_recommendation(
            StructuredAssessmentRecommendation(
                key='CUSTOMER_SCAN_RECOMMEND_URINARY_SYMPTOMS_STRUCTURED_ASSESSMENT',
                rank=1,
                button='Assess',
                patient=self.patient,
                questionnaires=[StructuredAssessmentQuestionnaireUrinarySymptoms],
                title='Assess Urinary Symptoms'))

        return result



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AcuteInpatient, AnnualWellnessVisit, CareServicesInLongTermResidentialFacility,
    DementiaMedications, Diabetes, Ed, EncounterInpatient, Ethnicity, FrailtyDevice,
    FrailtyDiagnosis, FrailtyEncounter, FrailtySymptom, Hba1CLaboratoryTest,
    HomeHealthcareServices, HospiceCareAmbulatory, NonacuteInpatient, NursingFacilityVisit,
    Observation, OfficeVisit, OncAdministrativeSex, Outpatient, Payer,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp, Race)


class ClinicalQualityMeasure122v9(ClinicalQualityMeasure):
    """
    Diabetes: Hemoglobin A1c (HbA1c) Poor Control (> 9%)

    Description: Percentage of patients 18-75 years of age with diabetes who had hemoglobin A1c >
    9.0% during the measurement period

    Definition: None

    Rationale: As the seventh leading cause of death in the U.S., diabetes kills approximately
    79,500 people a year and affects more than 30 million Americans (9.4 percent of the U.S.
    population) (Centers for Disease Control and Prevention [CDC], 2017a, 2017b). Diabetes is a
    long-lasting disease marked by high blood glucose levels, resulting from the body's inability
    to produce or use insulin properly (CDC, 2019). People with diabetes are at increased risk of
    serious health complications including vision loss, heart disease, stroke, kidney failure,
    amputation of toes, feet or legs, and premature death (CDC, 2016).

    In 2017, diabetes cost the U.S. an estimated $327 billion: $237 billion in direct medical costs
    and $90 billion in reduced productivity. This is a 34 percent increase from the estimated $245
    billion spent on diabetes in 2012 (American Diabetes Association, 2018).

    Controlling A1c blood levels help reduce the risk of microvascular complications (eye, kidney
    and nerve diseases) (CDC, 2014).

    Guidance: Patient is numerator compliant if most recent HbA1c level >9%, the most recent HbA1c
    result is missing, or if there are no HbA1c tests performed and results documented during the
    measurement period. If the HbA1c test result is in the medical record, the test can be used to
    determine numerator compliance.

    Only patients with a diagnosis of Type 1 or Type 2 diabetes should be included in the
    denominator of this measure; patients with a diagnosis of secondary diabetes due to another
    condition should not be included.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms122v9
    """

    title = 'Diabetes: Hemoglobin A1c (HbA1c) Poor Control (> 9%)'

    identifiers = ['CMS122v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'American Diabetes Association. (2018). Economic costs of diabetes in the U.S. in 2017. Diabetes Care, 41, 917-928. Retrieved from http://care.diabetesjournals.org/content/early/2018/03/20/dci18-0007',
        'American Diabetes Association. (2019). 6. Glycemic targets: Standards of medical care in diabetes—2019. Diabetes Care, 42(Suppl. 1), S61-S70. https://doi.org/10.2337/dc19-S006',
        'Centers for Disease Control and Prevention. (2014). National diabetes statistics report, 2014. Atlanta, GA: U.S. Department of Health and Human Services, CDC. Retrieved from http://www.thefdha.org/pdf/diabetes.pdf',
        'Centers for Disease Control and Prevention. (2016). At a glance 2016: Diabetes—Working to reverse the U.S. epidemic. Atlanta, GA: Author. Retrieved from https://upcap.org/admin/wp-content/uploads/2016/06/Diabetes-at-a-Glance.pdf',
        'Centers for Disease Control and Prevention. (2017a). Health, United States, 2016: With chartbook on long-term trends in health. Retrieved from https://www.cdc.gov/nchs/data/hus/hus16.pdf',
        'Centers for Disease Control and Prevention. (2017b). National diabetes statistics report, 2017. Atlanta, GA: U.S. Department of Health and Human Services, CDC. Retrieved from https://www.cdc.gov/diabetes/pdfs/data/statistics/national-diabetes-statistics-report.pdf',
        'Centers for Disease Control and Prevention. (2019). About diabetes. Retrieved from https://www.cdc.gov/diabetes/basics/diabetes.html',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients 18-75 years of age with diabetes with a visit during the
        measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients whose hospice care overlaps the measurement period.

        Exclude patients 66 and older who are living long term in an institution for more than 90
        consecutive days during the measurement period.

        Exclude patients 66 and older with advanced illness and frailty because it is unlikely that
        patients will benefit from the services being measured.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients whose most recent HbA1c level (performed during the measurement period)
        is >9.0%

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: American Diabetes Association (2019):

        - A reasonable A1C goal for many nonpregnant adults is <7%. (Level of evidence: A)

        - Providers might reasonably suggest more stringent A1C goals (such as <6.5% [48 mmol/mol])
        for selected individual patients if this can be achieved without significant hypoglycemia
        or other adverse effects of treatment (i.e. polypharmacy). Appropriate patients might
        include those with short duration of diabetes, type 2 diabetes treated with lifestyle or
        metformin only, long life expectancy, or no significant cardiovascular disease (CVD).
        (Level of evidence: C)

        - Less stringent A1C goals (such as <8% [64 mmol/mol]) may be appropriate for patients with
        a history of severe hypoglycemia, limited life expectancy, advanced microvascular or
        macrovascular complications, extensive comorbid
         conditions, or long-standing diabetes in whom the goal is difficult to achieve despite
        diabetes self-management education, appropriate glucose monitoring, and effective doses of
        multiple glucose-lowering agents including insulin. (Level of evidence: B)
        """
        pass



import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.cms131v6_diabetes_eye_exam import ClinicalQualityMeasure131v6
from canvas_workflow_kit.patient_recordset import BillingLineItemRecordSet, ProtocolOverrideRecordSet
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class ClinicalQualityMeasure131v6Test(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_enabled(self):
        self.assertTrue(ClinicalQualityMeasure131v6.enabled())

    def test_description(self):
        expected = ('Patients 18-75 years of age with diabetes who have not had a retinal or '
                    'dilated eye exam by an eye care professional.')
        self.assertEqual(expected, ClinicalQualityMeasure131v6._meta.description)

    def test_information(self):
        expected = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS131v6.html'
        self.assertEqual(expected, ClinicalQualityMeasure131v6._meta.information)

    def test_version(self):
        self.assertTrue(hasattr(ClinicalQualityMeasure131v6._meta, 'version'))

    def test_change_types(self):
        result = ClinicalQualityMeasure131v6._meta.compute_on_change_types
        expected = [
            'protocol_override',
            'condition',
            'patient',
            'referral_report',
        ]
        self.assertEqual(len(expected), len(result))
        for change in expected:
            self.assertIn(change, result)

    def test_in_denominator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has no diabetes --> false
        patient = self.load_patient('cms131v6_diabetes_no')
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient has diabetes --> true
        patient = self.load_patient('cms131v6_diabetes_yesnoreferralreport')
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient to young < 18 --> false
        patient = self.load_patient('cms131v6_diabetes_yesnoreferralreport')
        patient.patient['birthDate'] = '2000-08-24'
        self.assertTrue(17 < patient.age_at(timeframe.end) < 18)
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '2000-08-23'
        self.assertTrue(patient.age_at(timeframe.end) == 18)
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient to old > 75 --> false
        patient = self.load_patient('cms131v6_diabetes_yesnoreferralreport')
        patient.patient['birthDate'] = '1943-08-22'
        self.assertTrue(75 < patient.age_at(timeframe.end) < 76)
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '1943-08-23'
        self.assertTrue(patient.age_at(timeframe.end) == 75)
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '1943-08-24'
        self.assertTrue(patient.age_at(timeframe.end) < 75)
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient has an other code --> false
        patient = self.load_patient('cms131v6_diabetes_yesnoreferralreport')
        for condition in patient.conditions:
            for code in condition['coding']:
                code['code'] = 'code'
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient has diabetes + already perform the exam --> true
        patient = self.load_patient('cms131v6_diabetes_yeswithreferralreport')
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

    def test_in_numerator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # not test when patient has no diabetes because it makes no sense
        # patient = self.load_patient('cms131v6_diabetes_no')

        # patient has diabetes + no exam performed --> false
        patient = self.load_patient('cms131v6_diabetes_yesnoreferralreport')
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

        # patient has diabetes + already perform the exam, regardless the result --> true
        tests = [
            '37231002',
            '43959009',
            '4855003',
            '721103006',
        ]
        for on_date in ['2018-08-22', '2018-02-22', '2017-08-24']:
            for code in tests:
                patient = self.load_patient('cms131v6_diabetes_yeswithreferralreport')
                for reports in patient.referral_reports:
                    reports['originalDate'] = on_date
                    for coding in reports['codings']:
                        if coding['id'] == 59:
                            coding['code'] = code
                            break
                tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
                self.assertTrue(tested.in_numerator())

        # -- exams performed within the previous year of the measurement period
        # --> true for 721103006
        tests = [
            '37231002',
            '43959009',
            '4855003',
            '721103006',
        ]
        for on_date in ['2017-08-22', '2017-02-22', '2016-08-24']:
            for code in tests:
                patient = self.load_patient('cms131v6_diabetes_yeswithreferralreport')
                for reports in patient.referral_reports:
                    reports['originalDate'] = on_date
                    for coding in reports['codings']:
                        if coding['id'] == 59:
                            coding['code'] = code
                            break
                tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
                tested.in_numerator()

                if code == '721103006':
                    self.assertTrue(tested.in_numerator(), f'code {code} ({on_date})')
                else:
                    self.assertFalse(tested.in_numerator(), f'code {code} ({on_date})')

        # -- exams performed before the previous year of the measurement period --> false
        tests = [
            '37231002',
            '43959009',
            '4855003',
            '721103006',
        ]
        for on_date in ['2016-08-22', '2016-02-22']:
            for code in tests:
                patient = self.load_patient('cms131v6_diabetes_yeswithreferralreport')
                for reports in patient.referral_reports:
                    reports['originalDate'] = on_date
                    for coding in reports['codings']:
                        if coding['id'] == 59:
                            coding['code'] = code
                            break
                tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
                self.assertFalse(tested.in_numerator())

    def test_in_numerator_with_override(self):
        patient = self.load_patient('cms131v6_diabetes_yeswithreferralreport')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure131v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.referral_reports[0]['originalDate'] = '2018-08-05'

        tested = ClinicalQualityMeasure131v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=0))
        self.assertTrue(tested.in_numerator())

        tested = ClinicalQualityMeasure131v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=80))
        self.assertTrue(tested.in_numerator())

        tested = ClinicalQualityMeasure131v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=180))
        self.assertTrue(tested.in_numerator())

        tested = ClinicalQualityMeasure131v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=181))
        self.assertFalse(tested.in_numerator())

        # -- exams performed before the previous year of the measurement period --> false
        tests = [
            '37231002',
            '43959009',
            '4855003',
            '721103006',
        ]
        for code in tests:
            for reports in patient.referral_reports:
                reports['originalDate'] = '2018-08-05'
                for coding in reports['codings']:
                    if coding['id'] == 59:
                        coding['code'] = code
                        break
            tested = ClinicalQualityMeasure131v6(
                patient=patient, now=arrow.get('2018-08-05').shift(days=350))
            if code == '721103006':
                self.assertTrue(tested.in_numerator(), f'code {code}')
            else:
                self.assertFalse(tested.in_numerator(), f'code {code}')

            tested = ClinicalQualityMeasure131v6(
                patient=patient, now=arrow.get('2018-08-05').shift(days=361))
            self.assertFalse(tested.in_numerator(), f'code {code}')

    def test_compute_results(self):
        start = arrow.get('2017-09-03 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        now = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has no diabetes --> false
        patient = self.load_patient('cms131v6_diabetes_no')
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertIsNotApplicable(result)

        # patient has diabetes --> true
        patient = self.load_patient('cms131v6_diabetes_yesnoreferralreport')
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = ('Dohav has diabetes and no documentation of retinal '
                'examination in the past 12 months.')
        self.assertEqual(text, result.narrative)
        self.helper_exam_recommended(result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient has diabetes + already perform the exam --> true
        patient = self.load_patient('cms131v6_diabetes_yeswithreferralreport')
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = (r'^Dohav has diabetes and a retinal examination was done [\w\s]+ on 8/22/18,'
                ' demonstrating Retinopathy co-occurrent and due to diabetes mellitus.$')
        self.assertRegex(result.narrative, text)
        self.assertEqual([], result.recommendations)
        self.assertEqual(352, result.due_in)

        # -- exams performed within the previous year of the measurement period
        # --> true for 721103006
        tests = [
            '37231002',
            '43959009',
            '4855003',
            '721103006',
        ]
        for code in tests:
            patient = self.load_patient('cms131v6_diabetes_yeswithreferralreport')
            for reports in patient.referral_reports:
                reports['originalDate'] = '2017-02-22'
                for coding in reports['codings']:
                    if coding['id'] == 59:
                        coding['code'] = code
                        break
            tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe, now=now)
            result = tested.compute_results()

            if code == '721103006':
                text = (
                    r'^Dohav has diabetes and a retinal examination was done [\w\s]+ on 2/22/17 '
                    'demonstrating no diabetic eye disease. '
                    'Next examination is due 2/11/18.$')
                self.assertRegex(result.narrative, text)
                self.assertEqual([], result.recommendations)
                self.assertEqual('satisfied', result.status)
                self.assertEqual(-194, result.due_in)
            else:
                text = (
                    r'^Dohav has diabetes and a prior abnormal retinal examination [\w\s]+ '
                    'on 2/22/17 showing Retinopathy co-occurrent and due to diabetes mellitus. '
                    'Dohav is due for retinal examination.$')
                self.assertRegex(result.narrative, text)
                self.helper_exam_recommended(result.recommendations)
                self.assertEqual('due', result.status)
                self.assertEqual(-1, result.due_in)

        # -- exams performed before the previous year of the measurement period --> false
        tests = [
            '37231002',
            '43959009',
            '4855003',
            '721103006',
        ]
        for on_date in ['2016-08-22', '2016-02-22']:
            for code in tests:
                patient = self.load_patient('cms131v6_diabetes_yeswithreferralreport')
                for reports in patient.referral_reports:
                    reports['originalDate'] = on_date
                    for coding in reports['codings']:
                        if coding['id'] == 59:
                            coding['code'] = code
                            break
                tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe, now=now)
                result = tested.compute_results()
                text = ('Dohav has diabetes and no documentation of retinal '
                        'examination in the past 12 months.')
                self.assertEqual(text, result.narrative)
                self.helper_exam_recommended(result.recommendations)
                self.assertEqual('due', result.status)
                self.assertEqual(-1, result.due_in)

        # patient has diabetes + younger than 18 --> n/a + due in
        patient = self.load_patient('cms131v6_diabetes_no')
        patient.patient['birthDate'] = '2000-08-30'
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertIsNone(result.due_in)
        patient = self.load_patient('cms131v6_diabetes_yesnoreferralreport')
        patient.patient['birthDate'] = '2000-08-30'
        tested = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual(6, result.due_in)

    def test_compute_results_with_override(self):
        patient = self.load_patient('cms131v6_diabetes_yeswithreferralreport')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure131v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.referral_reports[0]['originalDate'] = '2018-08-05'

        # exam done within the last 180 days
        tested = ClinicalQualityMeasure131v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=80))
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = ('Dohav has diabetes and a retinal examination was done 2 months ago on 8/5/18, '
                'demonstrating Retinopathy co-occurrent and due to diabetes mellitus.')
        self.assertEqual(text, result.narrative)
        self.assertEqual(0, len(result.recommendations))
        self.assertEqual(100, result.due_in)

        # exam done 181 days
        patient.billing_line_items = BillingLineItemRecordSet([{
            'id': 92,
            'created': arrow.get('2018-08-01').shift(days=181),
            'datetimeOfService': arrow.get('2018-08-01').shift(days=181),
            'cpt': '99202',
            'units': 1
        }])
        tested = ClinicalQualityMeasure131v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=181))
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = (
            'Dohav has diabetes and a prior abnormal retinal examination 6 months ago on 8/5/18 '
            'showing Retinopathy co-occurrent and due to diabetes mellitus. '
            'Dohav is due for retinal examination.')
        self.assertEqual(text, result.narrative)
        self.assertEqual(2, len(result.recommendations))
        self.assertEqual(-1, result.due_in)

        # exam done 240 days ago
        tests = [
            '37231002',
            '43959009',
            '4855003',
            '721103006',
        ]
        for code in tests:
            for reports in patient.referral_reports:
                reports['originalDate'] = '2018-08-05'
                for coding in reports['codings']:
                    if coding['id'] == 59:
                        coding['code'] = code
                        break
            tested = ClinicalQualityMeasure131v6(
                patient=patient, now=arrow.get('2018-08-05').shift(days=240))
            result = tested.compute_results()

            if code == '721103006':
                text = ('Dohav has diabetes and a retinal examination was done 8 months ago '
                        'on 8/5/18 demonstrating no diabetic eye disease. '
                        'Next examination is due 2/1/19.')
                self.assertEqual(text, result.narrative)
                self.assertEqual([], result.recommendations)
                self.assertEqual('satisfied', result.status)
                self.assertEqual(-60, result.due_in)
            else:
                text = ('Dohav has diabetes and a prior abnormal retinal examination 8 months ago '
                        'on 8/5/18 showing Retinopathy co-occurrent and due to diabetes mellitus. '
                        'Dohav is due for retinal examination.')
                self.assertEqual(text, result.narrative)
                self.helper_exam_recommended(result.recommendations)
                self.assertEqual('due', result.status)
                self.assertEqual(-1, result.due_in)

    def helper_exam_recommended(self, result):
        expected = [
            {
                'title': 'Perform retinal examination',
                'narrative': None,
                'command': {
                    'type': 'perform',
                    'filter': {
                        'coding': [{
                            'system': 'cpt',
                            'code': ['92250']
                        }]
                    }
                },
                'key': 'CMS131v6_RECOMMEND_PERFORM_EYE_EXAM',
                'rank': 1,
                'button': 'Perform',
            },
            {
                'title': 'Refer for retinal examination',
                'narrative': None,
                'command': {
                    'type': 'refer',
                    'filter': {
                        'coding': [{
                            'system': 'snomedct',
                            'code': [
                                '427478009', '308110009', '410453006', '314972008', '410451008',
                                '252789008', '410452001', '252784003', '274798009', '314971001',
                                '410455004', '420213007', '252780007', '6615001', '425816006',
                                '252790004', '252782004', '274795007', '252788000', '252779009',
                                '252781006', '252783009'
                            ]
                        }]
                    }
                },
                'key': 'CMS131v6_RECOMMEND_REFER_EYE_EXAM',
                'rank': 2,
                'button': 'Refer',
            },
        ]

        for exp, recommendation in zip(expected, result):
            self.assertEqual(exp['key'], recommendation.key)
            self.assertEqual(exp['rank'], recommendation.rank)
            self.assertEqual(exp['button'], recommendation.button)
            self.assertEqual(exp['title'], recommendation.title)
            self.assertEqual(exp['narrative'], recommendation.narrative)
            self.assertEqual(exp['command']['type'], recommendation.command['type'])
            exp_item = exp['command']['filter']['coding']
            rec_item = recommendation.command['filter']['coding']
            self.assertEqual(len(exp_item), len(rec_item))
            for exp_coding, rec_coding in zip(exp_item, rec_item):
                self.assertEqual(exp_coding['system'], rec_coding['system'])
                self.assertEqual(sorted(exp_coding['code']), sorted(rec_coding['code']))
